#!/usr/bin/python
"This module provides tabulated properties of the examined material"
import os
import sys
from abc import ABCMeta, abstractmethod
from math import acos, cos, exp, expm1, fabs, pi, radians, sin, sqrt
from re import sub as filt
from scipy import interpolate
from scipy.constants import N_A
import numpy as np

class Element(object):
    "Defines an element/isotope used for diffraction"
    __metaclass__ = ABCMeta
    _red_avogadro = 0.60221415 #Avogadro number in 10^24 units
    def __init__(self, name_par, mass_par=0.0, murho_par=0.0):
        "Sets name of element of the isotope, its mass and mu/rho value"
        self.name = name_par.upper()    #uppercase name of the element
        self.mass = mass_par            #mass of the element
        self.murho = murho_par          #mu/rho value belong to the element
        return

    @abstractmethod
    def get_f(self, q_value=0.0):
        r"Returns with the f(Q) value in Re and Im, if q_value provided in \AA-1 units"
        pass

    @abstractmethod
    def get_inelastic(self, q_value=0.0):
        r"Returns with the Compton/inelastic scattering factor, if q_value provided in \AA-1 units"
        pass

    def get_incoherent_intrinsic(self, q_vale=0.0):
        "Returns with the incoherent scattering cross-section/4*pi belonging the given element\
        in square of units used"
        return 0.0

    def get_murho(self, q_vale=0.0):
        "Returns with the mu/rho (mass absorption coefficient) in cm^2/g units"
        return self.murho

class NElement(Element):
    "Defines elements or isotopes used at neutron diffraction"
    def __init__(self, isotope_name, energy_or_wl_string, tables_path="tables"):
        "Initializes the structure as isotope name and string containing either wavelength or\
        energy indicating the type by correct dimension: A, m/s, eV or meV. Finally, path of\
        library containing the tables (without / at ind) should be provided"
        ener = energy_or_wl_string.strip().upper()
        #energy_or_wl_string.uppercase()
        if ener[-3:] == 'MEV':
            wavel = self.mev2a(float(energy_or_wl_string[0:-3]))
        elif ener[-3:] == 'M/S':
            wavel = self.mps2a(float(energy_or_wl_string[0:-3]))
        elif ener[-2:] == 'EV':
            wavel = self.ev2a(float(energy_or_wl_string[0:-2]))
        elif ener[-1] == 'A':
            wavel = float(energy_or_wl_string[0:-1])
        else:
            print 'You entered "', energy_or_wl_string,\
                '" does not correspond to any A or EV or mEV dimension indicator at the end'
            quit()
            return
        #initialize variables
        super(NElement, self).__init__(isotope_name)
        self.coh_scatt_length = np.zeros(2) #Coherent scattering length(fm): Re and Img parts
        self.incoherent_xs = 0.0            #incoherent scattering cross-section in barn units
        if self.__read_neutron_table(tables_path+"/nparm.txt") != 0:
            print 'The NElement of "'+isotope_name+'" is not correctly initialized!'
        else:
            self.__update_data_atenergy(self.a2mps(wavel))
            self.murho += (self.incoherent_xs+\
                (self.coh_scatt_length[0]**2+self.coh_scatt_length[1]**2)/25.0*pi)/self.mass*\
                    Element._red_avogadro
        return

    #Unit conversion for neutrons
    @staticmethod
    def ev2a(energy):
        r"eV->\AA conversion"
        return sqrt(0.08180420961/energy)

    @staticmethod
    def a2ev(wavl):
        r"\AA->eV conversion"
        return 0.08180420961/wavl**2

    @staticmethod
    def mev2a(energy):
        r"meV->\AA conversion"
        return NElement.ev2a(0.001*energy)

    @staticmethod
    def a2mev(wavl):
        r"\AA->keV conversion"
        return NElement.ev2a(wavl)*1000.0

    @staticmethod
    def mps2a(velocity):
        r"m/s->\AA conversion"
        return 3956.033568/velocity

    @staticmethod
    def a2mps(wavl):
        r"\AA->m/s conversion"
        return NElement.mps2a(wavl)

    def get_f(self, q_value=0.0):
        r"Returns with the b(Q) value in Re and Im, if q_value provided in \AA-1 units"
        return self.coh_scatt_length

    def get_inelastic(self, q_value=0.0):
        r"Returns with the inelastic scattering factor, if q_value provided in \AA-1 units"
        print "WARNING: Inelastic part of neutron has not been implemented yet!"
        return 0.0

    def get_incoherent_intrinsic(self, q_vale=0.0):
        "Returns with the incoherent scattering cross-section/4*pi belonging the given isotope\
        in square of units used"
        return self.incoherent_xs*25./pi

    def __read_neutron_table(self, full_filename):
        "Sets the coherent scattering length, the incoherent cross-section, the mu/rho and mass of\
        the isotope from neutron datatable. The full_filename file containing the previously\
        mentioned datasets (Sears tables) for self.name (might be element for natural\
        abundance -- example: hydrogen element: H, deuterium 2H). Returns 0 with success or -1 if\
        file do not exist, -2 if element is not on the list"
        found = False
        if not os.path.isfile(full_filename):
            print 'ERROR(NElement::__read_neutron_table): the "', full_filename,\
                '" file do not exist!'
            return -1
        else:
            with open(full_filename, 'r') as fil:
                for line in fil:
                    lin = line.strip().split()
                    if self.name == lin[0]:
                        found = True
                        self.mass = float(lin[2])
                        coherent = complex(lin[3].replace('I', 'j'))
                        self.coh_scatt_length = np.array([float(coherent.real),\
                            float(coherent.imag)])
                        self.incoherent_xs = float(lin[6])
                        self.murho = float(lin[8])/self.mass*Element._red_avogadro
                        break
            if not found:
                print 'ERROR(NElement::__read_neutron_table) the "', full_filename,\
                    '" file does not have any "', self.name, '" entry!'
                return -2
            else:
                return 0

    def __update_data_atenergy(self, present_speed, reference_speed=2200.0):
        "Updates the cross-section data at present speed if reference speed is provided in m/s"
        self.murho *= reference_speed/present_speed
        return

class XElement(Element):
    "Defines elements used at for x-ray diffraction"
    #Path for x-ray data tables
    FTAB_PATH = "xparm_ff.txt"
    #File path of Waasmaier&Kirfel tables containing coefficients of f
    COMP_PATH = "xparm_ci.txt"
    #File path of Balyuzi tables containing coefficients of compton intensities

    def __init__(self, element_name, energy_or_wl_string, tables_path="tables"):
        "Initializes the structure as element name and string containing either wavelength or\
        energy indicating the type by correct dimension: A, eV or keV. Finally, path of library\
        containing the tables (without / at ind) should be provided"
        ener = energy_or_wl_string.strip().upper()
        #energy_or_wl_string.uppercase()
        if ener[-3:] == 'KEV':
            wavel = self.kev2a(float(energy_or_wl_string[0:-3]))
        elif ener[-2:] == 'EV':
            wavel = self.ev2a(float(energy_or_wl_string[0:-2]))
        elif ener[-1] == 'A':
            wavel = float(energy_or_wl_string[0:-1])
        else:
            print 'You entered "', energy_or_wl_string,\
                '" does not correspond to any A or EV or KEV dimension indicator at the end'
            quit()
            return
        #initialize variables
        super(XElement, self).__init__(element_name)
        self.f_scatt = np.zeros(12)         #contains the coefficients of f in 5{a,b},c,f" order
        self.compton = np.zeros(11)         #Compton scattering coefficients in 5{a,b},Zeff
        retval = self.__get_absorption_murho_mass(filt('[^A-Z]', '', self.name), wavel,\
            tables_path+'/x_abs_sasaki/')
        if retval == 0:
            retval = self.__get_f_compton(tables_path+'/xparm_ff.txt', self.name, True)
        if retval == 0:
            retval = self.__get_f_compton(tables_path+'/xparm_ci.txt', self.name, False)
        #if retval == 0:
        #    retval = self.__get_fprime_dprime(filt('[^A-Z]', '', self.name), wavel,\
        #    tables_path+'/x_anom_sasaki/')
        if retval != 0:
            print 'The XElement of "'+element_name+'" is not correctly initialized!'
        return

    #Unit conversion for photons
    @staticmethod
    def ev2a(energy):
        r"eV->\AA conversion"
        return 12398.42/energy

    @staticmethod
    def a2ev(wavl):
        r"\AA->eV conversion"
        return XElement.ev2a(wavl)

    @staticmethod
    def kev2a(energy):
        r"keV->\AA conversion"
        return XElement.ev2a(1000.0*energy)

    @staticmethod
    def a2kev(wavl):
        r"\AA->keV conversion"
        return XElement.ev2a(wavl)/1000.0

    def get_f(self, q_value=0.0):
        r"Returns with the f(Q) value in Re and Im, if q_value provided in \AA-1 units"
        re_part = 0.0
        for i in range(5):
            re_part += self.f_scatt[2*i] * exp(-self.f_scatt[2*i+1] * q_value**2 / (16.0 * pi**2))
        return re_part + self.f_scatt[-2], self.f_scatt[-1]

    def get_inelastic(self, q_value=0.0):
        r"Returns with the Compton scattering factor, if q_value provided in \AA-1 units"
        re_part = 0.0
        for i in range(5):
            re_part += self.compton[2*i] * exp(-self.compton[2*i+1] * q_value**2 / (16.0 * pi**2))
        return self.compton[-1] - re_part

    @staticmethod
    def _read_sasaki_file(filename, xcolumn, ycolumns, xvalue, yvals):
        "Read a sasakilab kind of file specified by filename, containing the x datasets in xcolumn,\
        the y dataset(s) in ycolumns. It determines the y values belonging to xvalue and appends to\
        yvals. It returns with zero in the case of success; -1 if file does not exist and -2 if\
        xvalue is below, -3 if over the valid xrange"
        if not os.path.isfile(filename):
            return -1
        cols = xcolumn,
        if isinstance(ycolumns, tuple):
            cols += ycolumns
        else:
            cols += ycolumns,
        data = np.loadtxt(filename, usecols=cols)
        if data[0, 0] > xvalue and data[-1, 0] > xvalue:
            return -2
        if data[0, 0] < xvalue and data[-1, 0] < xvalue:
            return -3
        for i in range(1, len(data[0, :])):
            #interpolating the calculated data
            tck = interpolate.splrep(data[:, 0], data[:, i], s=0)
            yvals.append(interpolate.splev(xvalue, tck, der=0))
        return 0

    def __get_absorption_murho_mass(self, element_pure_uppercase, wavelength, abspath):
        "Gets the mu/rho and mass of the element provided by element_pure_uppercase. It uses\
        tables,which are appearing in abspath library in NAME~X.txt using Sasakilab format,\
        where X=w, k,l1,l2,l3 are considered. Wavelength is necessary to provide the right values"
        if element_pure_uppercase == 'H' or element_pure_uppercase == 'HA' or\
            element_pure_uppercase == 'HS':
            print 'WARNING(XElement::__getAbsorptionMuRhoMass): file for hidrogen does not exist!'
            print '\tmu/rho=0 and the correct mass has been set up!'
            self.mass = 1.00794
            return 0
        elif element_pure_uppercase == 'HE':
            print 'WARNING(XElement::__getAbsorptionMuRhoMass): file for helium does not exist!'
            print '\tmu/rho=0 and the correct mass has been set up!'
            self.mass = 4.002602
            return 0
        elif not os.path.isfile(abspath+element_pure_uppercase+'~w.txt'):
            print 'ERROR(XElement::__getAbsorptionMuRhoMass): the "', abspath,\
                element_pure_uppercase, '~w.txt" file do not exist!'
            return -1
        else:
            #get the mass
            with open(abspath+element_pure_uppercase+'~w.txt', 'r') as fil:
                #print 'Looking for mass in file "', abspath+element_pure_uppercase, '~w.txt"...'
                self.mass = float(fil.readline().split()[12])
            #get the mu/rho
            actual = []
            retval = self._read_sasaki_file(abspath+element_pure_uppercase+'~w.txt', 0, 2,\
                wavelength, actual)
            if retval < -1:
                print "The requested x value is outside the valid energy/wavelength range!"
                quit()
                return retval
            else:
                self.murho = actual[0]
                #We have one valid data, check other files, wl might be close to absorption edge
                for fnam in 'k', 'l1', 'l2', 'l3':
                    act = []
                    retval = self._read_sasaki_file(abspath+element_pure_uppercase+'~'+fnam+\
                        '.txt', 0, 2, wavelength, act)
                    if retval == 0:
                        self.murho = act[0]
            return 0

    def __get_f_compton(self, full_filename, element_uppercase, ft_comptonf):
        "Sets the f_scatt (True) or compton (ft_comptonf = False) coefficients from full_filename\
        file containing the full path of Waasmaier&Kirfel or Balyuzzi tables for element_uppercase\
        having the correct charge. Returns 0 with success or -1 if file do not exist, -2 if\
        element is not on the list"
        found = False
        if not os.path.isfile(full_filename):
            print 'ERROR(XElement::__get_f_compton): the "', full_filename, '" file do not exist!'
            return -1
        else:
            element = filt('[^A-Z]', '', element_uppercase) #supposing, that the same intensity\
                                                            #will be used for different each ions
            with open(full_filename, 'r') as fil:
                for line in fil:
                    lin = line.strip().split()
                    if (ft_comptonf and element_uppercase == lin[0]) or\
                        ((not ft_comptonf) and element == lin[0]):
                        found = True
                        if ft_comptonf:
                            for i in range(1, len(self.f_scatt)):
                                self.f_scatt[i-1] = lin[i]
                        else:
                            for i in range(1, len(self.compton)+1):
                                self.compton[i-1] = lin[i]
                            remaining = element_uppercase[element_uppercase.find(lin[0])+\
                                len(lin[0]):]
                            if remaining.find('-') != -1:
                                idx = remaining.find('-')
                                self.compton[10] += float(remaining[0:idx])
                            elif remaining.find('+') != -1:
                                idx = remaining.find('+')
                                self.compton[10] -= float(remaining[0:idx])
                        break
            if not found:
                print 'ERROR(XElement::__get_f_compton): the "', full_filename,\
                    '" file does not have any "', element_uppercase, '" entry!'
                return -2
            else:
                return 0

    def __get_fprime_dprime(self, element_pure_uppercase, wavelength, anompath):
        "Gets f' and f\" anomalous contributions for element provided by element_pure_uppercase.\
        It uses tables,which are appearing in anompath library in NAME~X.txt using Sasakilab\
        format, where X=w, k,l1,l2,l3 are considered. Wavelength is necessary to provide the right\
        values"
        if element_pure_uppercase == 'H' or element_pure_uppercase == 'HA' or\
            element_pure_uppercase == 'HS':
            print 'WARNING(XElement::__get_fprime_dprime): file for hidrogen does not exist!'
            print '\tthe f\'=0 and f"=0 has been supposed!'
            return 0
        elif element_pure_uppercase == 'HE':
            print 'WARNING(XElement::__get_fprime_dprime): file for helium does not exist!'
            print '\tthe f\'=0 and f"=0 has been supposed!'
            return 0
        elif not os.path.isfile(anompath+element_pure_uppercase+'~w.txt'):
            print 'ERROR(XElement::__get_fprime_dprime): the "', anompath,\
                element_pure_uppercase, '~w.txt" file do not exist!'
            return -1
        else:
            #get the f' and f"
            fprime = 0.0
            fdprime = 0.0
            actual = []
            retval = self._read_sasaki_file(anompath+element_pure_uppercase+'~w.txt', 0, (2, 3),\
                wavelength, actual)
            if retval < -1:
                print "The requested x value is outside the valid energy/wavelength range!"
                quit()
                return retval
            else:
                fprime = actual[0]
                fdprime = actual[1]
                #We have one valid data, check other files, wl might be close to absorption edge
                for fnam in 'k', 'l1', 'l2', 'l3':
                    act = []
                    retval = self._read_sasaki_file(anompath+element_pure_uppercase+'~'+fnam+\
                        '.txt', 0, 2, wavelength, act)
                    if retval == 0:
                        fprime = act[0]
                        fdprime = act[1]
                self.f_scatt[-2] += fprime
                self.f_scatt[-1] += fdprime
            return 0

class Material(object):
    "This class describes a material composed by elements"
    _red_avogadro = 0.60221415 #Avogadro number in 10^24 units
    def __init__(self, density, fullness=1.0):
        "Initializes the data structure and sets the density provided in g/cm^3 and fullness\
        parameter to model porosity"
        self.elements = []
        self.contribution = []
        self._density = density
        self.fullness = fullness

    def add_element(self, element_obj, percent):
        "Add element to the material contains the atomic % of the material"
        self.elements.append(element_obj)
        self.contribution.append(percent)
        return

    def get_mu_rho(self):
        "Returns with mu/rho in cm^2/g units"
        #sum_percentage = sum(self.contribution)
        sum_weight = sum(self.elements[i].mass * contrib\
            for i, contrib in enumerate(self.contribution))
        sum_weight_mu = sum(self.elements[i].mass * self.contribution[i] * self.elements[i].murho\
            for i in range(len(self.contribution)))
        return sum_weight_mu / sum_weight * self.fullness

    def get_lin_abs_coeff(self):
        "Returns with the linear absorption coefficient of the material in 1/cm units"
        return self.get_mu_rho()*self._density

    def get_number_density(self):
        r"Returns with the number density in 1/\AA**3"
        sum_weight = sum(self.elements[i].mass * sc\
            for i, sc in enumerate(self.contribution))
        sum_perc = sum(self.contribution)
        return self._density / sum_weight * self._red_avogadro * sum_perc

    def get_density(self):
        "Returns with the value of density in g/cm**3"
        return self._density

    def set_number_density(self, ndens):
        "This method sets the number density"
        sum_weight = sum(self.elements[i].mass * sc\
            for i, sc in enumerate(self.contribution))
        sum_perc = sum(self.contribution)
        self._density = ndens / self._red_avogadro / sum_perc * sum_weight
        return

    def get_f_aver(self, q_value=0.):
        "Returns with the Re and Im part of <f(Q)> value"
        sumr = 0.
        sumi = 0.
        for i in range(len(self.contribution)):
            result = self.elements[i].get_f(q_value)
            sumr += result[0]*self.contribution[i]
            sumi += result[1]*self.contribution[i]
        sum_perc = sum(self.contribution)
        return sumr/sum_perc, sumi/sum_perc

    def get_f2_aver(self, q_value=0.):
        "Returns with the <f^2(Q)> value"
        sumr = 0.
        for i in range(len(self.contribution)):
            result = self.elements[i].get_f(q_value)
            sumr += (result[0]**2+ result[1]**2)*self.contribution[i]
        sum_perc = sum(self.contribution)
        return sumr/sum_perc

    def get_f_aver2(self, q_value=0.):
        "Returns with the <f(Q)>^2 value"
        result = self.get_f_aver(q_value)
        return result[0]**2+result[1]**2

    def get_coherent(self, q_value=0.):
        "Returns with the <f>^2"
        return self.get_f_aver2(q_value)

    def get_incoherent_intrinsic(self, q_value=0.):
        "Returns with the intrinsic incoherent contribution"
        intrinsic = 0.
        for i in range(len(self.contribution)):
            intrinsic += self.elements[i].get_incoherent_intrinsic(q_value)*self.contribution[i]
        return intrinsic/sum(self.contribution)

    def get_incoherent(self, q_value=0.):
        "Returns with <f^2>-<f>^2 and with the intrinsic incoherent scattering contribution"
        return self.get_f2_aver(q_value) - self.get_f_aver2(q_value) +\
            self.get_incoherent_intrinsic(q_value)

    def get_inelastic(self, q_value=0.):
        "Returns with the Compton scattering coefficient"
        sum_perc = sum(self.contribution)
        return sum(self.elements[i].get_inelastic(q_value) * self.contribution[i]\
            for i in range(len(self.contribution)))/ sum_perc

class Geometry(object):
    "This class describes the geometry from inside to outside in the case of cylindrical and\
    from first hit to last hit order in flat plate geometry"
    __metaclass__ = ABCMeta
    def __init__(self, lin_sizes, num_dens, lin_abs):
        "Specifies the geometry: the lin_sizes are linear sizes (thickness for flat plate and\
        radius in cylindrical) in mm, number densities in atom/A^3 and lin_abs are the linear\
        absorption coefficient in 1/cm units. Each value should be a numpy array with same\
        length, corresponding the number of layers/cylinders"
        if len(lin_sizes) == len(num_dens) and len(num_dens) == len(lin_abs):
            self.thickness = lin_sizes/10.0             #thickness in cm units of each layers
            self.num_dens = num_dens*N_A*1.0e-24   #number density in mol/cm^3 units
            self.labs = lin_abs                         #linear absorption coefficient in 1/cm
            return
        else:
            print "ERROR:(Geometry): The length of each input arrays should be the same!"
        return

    @abstractmethod
    def get_number_of_scatterers(self, beam_width, beam_height, layer_indices, theta_incoming):
        "Returns with the number of scatterers in mole units. Inputs are beam profile are in mm,\
        layer indices of the sample starting from zero from the direction of source and incoming\
        theta degree"
        pass

    @abstractmethod
    def calc_single_scattering(self, layer_contributes, layer_scattering, theta_incoming,\
        theta_outgoing=None):
        "Calculates the transmission coefficients belonging to layer_scattering layers, if\
        layer_contributes layers arepresent (these two are boolean arrays) at theta_incoming\
        and theta outgoing degrees. If outgoing degrees are not provided, symmetrical condition\
        is supposed"
        pass

class FlatPlateGeometry(Geometry):
    "This class describes flat plate geometry from source layer to furthest layer.\
    Here, sample bigger than beam condition is supposed."
    @staticmethod
    def __f(value):
        "This function calculates the (1-e^(-x))/x expression"
        if fabs(value) <= 0.1:
            return 1.0-0.5*value+value**2/6.-value**3/24.+value**4/120.0
        else:
            return -expm1(-value)/value

    def flat_refl(self, optical_pathlength, layer_contributes, layer_scattered):
        "This function calculates the transmission coefficient for reflection geometry.\
        The optical pathlength contain the effective optical pathlength for each layer,\
        layer_contributes and layer_scattered are boolean arrays. The former indicate its\
        contribution at absorption calculation, while the least shows those layers, whose\
        scattering transmission should be  calculated"
        nominator = 0.0
        denominator = 0.0
        for j in range(len(self.labs)):
            if layer_scattered[j]:
                if not layer_contributes[j]:
                    print 'ERROR(FlatPlateGeometry.flat_refl): the layer no. ', j,\
                        ' belongs to scatterers, but do not appear in contributing layers!'
                    quit()
                nominator += self.num_dens[j]*self.thickness[j]
                expsum = 0.0
                for i in range(j):
                    if layer_contributes[i]:
                        expsum += optical_pathlength[i]
                denominator += self.__f(optical_pathlength[j])*self.num_dens[j]*self.thickness[j]\
                    *exp(-expsum)
        return denominator/nominator

    def get_number_of_scatterers(self, beam_width, beam_height, layer_indices, theta_incoming=None):
        "Returns with the number of scatterers in mole units. Inputs are beam profile are in mm,\
        layer indices of the sample starting from zero from the direction of source and incoming\
        theta degree"
        if theta_incoming is None:
            result = 0.0
            for i in layer_indices:
                result += self.thickness[i]*self.num_dens[i]
            return result*beam_width*beam_height/100.0+(theta_incoming-theta_incoming)
        else:
            return -1.0

    @abstractmethod
    def calc_single_scattering(self, layer_contributes, layer_scattering, theta_incoming,\
        theta_outgoing=None):
        "Calculates the transmission coefficients belonging to layer_scattering layers, if\
        layer_contributes layers arepresent (these two are boolean arrays) at theta_incoming\
        and theta outgoing degrees. If outgoing degrees are not provided, symmetrical condition\
        is supposed"
        pass

class FlatPlateGeometryTransmission(FlatPlateGeometry):
    "This class describes transmission flat plate geometry from source layer to furthest layer.\
    Here, sample bigger than beam condition is supposed. Angles are measured perpendicular to the\
    plane of the scattering object."
    def get_number_of_scatterers(self, beam_width, beam_height, layer_indices, theta_incoming=None):
        "Returns with the number of scatterers in mole units. Inputs are beam profile are in mm,\
        layer indices of the sample starting from zero from the direction of source and incoming\
        theta degree"
        if theta_incoming is None:
            return -1
        else:
            return FlatPlateGeometry.get_number_of_scatterers(beam_width, beam_height,\
                layer_indices)/cos(radians(theta_incoming))

    def calc_single_scattering(self, layer_contributes, layer_scattering, theta_incoming,\
        theta_outgoing=None):
        "Calculates the transmission coefficients belonging to layer_scattering layers, if\
        layer_contributes layers arepresent (these two are boolean arrays) at theta_incoming\
        and theta outgoing degrees. If outgoing degrees are not provided, symmetrical condition\
        is supposed"
        if theta_outgoing is None:
            theta_outgoing = theta_incoming
        alpha = cos(radians(theta_incoming))
        beta = cos(radians(theta_outgoing))
        optical_pathlength_difference = np.zeros(len(self.labs))
        path = 0.0
        for i in range(len(self.labs)):
            optical_pathlength_difference[i] = self.labs[i] * self.thickness[i] * (beta - alpha)\
                /(alpha*beta)
            if layer_contributes[i]:
                path += self.labs[i] * self.thickness[i]/beta
        return exp(-path)*self.flat_refl(optical_pathlength_difference, layer_contributes,\
            layer_scattering)

class FlatPlateGeometryReflection(FlatPlateGeometry):
    "This class describes transmission flat plate geometry from source layer to furthest layer.\
    Here, sample bigger than beam condition is supposed. Angles are measured parallel to the\
    plane of the scattering object."
    def get_number_of_scatterers(self, beam_width, beam_height, layer_indices, theta_incoming=None):
        "Returns with the number of scatterers in mole units. Inputs are beam profile are in mm,\
        layer indices of the sample starting from zero from the direction of source and incoming\
        theta degree"
        if theta_incoming is None:
            return -1
        else:
            return FlatPlateGeometry.get_number_of_scatterers(beam_width, beam_height,\
                layer_indices)/sin(radians(theta_incoming))

    def calc_single_scattering(self, layer_contributes, layer_scattering, theta_incoming,\
        theta_outgoing=None):
        "Calculates the transmission coefficients belonging to layer_scattering layers, if\
        layer_contributes layers arepresent (these two are boolean arrays) at theta_incoming\
        and theta outgoing degrees. If outgoing degrees are not provided, symmetrical condition\
        is supposed"
        if theta_outgoing is None:
            theta_outgoing = theta_incoming
        alpha = sin(radians(theta_incoming))
        beta = sin(radians(theta_outgoing))
        optical_pathlength_difference = np.zeros(len(self.labs))
        for i, lcon in enumerate(layer_contributes):
            if lcon:
                optical_pathlength_difference[i] = self.labs[i] * self.thickness[i]\
                    * (beta + alpha)\
                /(alpha*beta)
        return self.flat_refl(optical_pathlength_difference, layer_contributes, layer_scattering)

class CylindricalGeometry(Geometry):
    "This class describes cylindrical geometry from inner cylinder to outer cylinder direction.\
    Here, beam bigger than cylinder condition is supposed."
    def __init__(self, lin_sizes, num_dens, lin_abs, calc_rings_angle_base_segments=None):
        "Specifies the geometry: the lin_sizes are linear sizes (thickness for flat plate and\
        radius in cylindrical) in mm, number densities in atom/A^3 and lin_abs are the linear\
        absorption coefficient in 1/cm units. Finally, the calc_rings_angle_base_segments\
        defines the number of calculation rings and base segments for each rings in (#rings,#base)\
        order. Eahc value should be a numpy array with same length, corresponding the number of\
        cylinders"
        Geometry.__init__(self, lin_sizes, num_dens, lin_abs)
        self.rsquares = self.thickness**2
        if calc_rings_angle_base_segments is None:
            #The following experimental formula deduced by fitting an y=A*x**B+C kind of expression
            #to the transmission coefficients belonging to each rings. These are determined on the
            #basis of looking for the maximal muR by tolerating 0.5% relative difference in
            #transmission coefficient to the reference system (500 rings, 11 segments) under
            #keeping the number of rings fixed.
            no_rings = np.ceil(6.82423*(self.thickness*self.labs)**2.11432+5.55588)
            self.no_inner_rings = no_rings.astype(int)
            #The previously mentioned investigation clarified, that the number of rings has a minor
            #effect for the value of the coefficients, except for low ring numbers
            self.start_segments = np.full(len(self.thickness), 3, dtype=int)
        else:
            self.no_inner_rings = calc_rings_angle_base_segments[::2]
            self.start_segments = calc_rings_angle_base_segments[1::2]
        return

    @staticmethod
    def __circle_area(radius, width):
        "Calculates the area of that part of a circle having a proper radius, which is less than\
        +/-width/2 to the central line"
        if width/2 > radius:
            return radius**2*pi
        else:
            rate = width/2./radius
            return radius**2*(pi-2*(acos(rate)+rate*sqrt(1.0-rate**2)))

    def __calculate_optical_path(self, scattering_ring, point, ring_present,\
        incomingt_outgoingf):
        "This method calculates the optical path to point(x,y) for incoming(t) and from\
        in the case of outgoing(f) path. The scattering point is situated in scattering_ring, and\
        the presence of rings is set by ring_present boolean array"
        ysq = point[1]**2
        opath = 0.0
        for i in range(scattering_ring+1, len(self.labs)):
            #until the scattering ring sum all of contributions
            if ring_present[i]:
                opath += self.labs[i]*(sqrt(self.rsquares[i]-ysq)-sqrt(self.rsquares[i-1]-ysq))
        chance2innerring = False
        if (incomingt_outgoingf and point[0] > 0) or ((not incomingt_outgoingf) and point[0] < 0):
            chance2innerring = True
        if scattering_ring > 0 and self.rsquares[scattering_ring-1] > ysq and chance2innerring:
            #the path pass through the domain of inner rings -- add the term of scattering ring
            opath += self.labs[scattering_ring]*(sqrt(self.rsquares[scattering_ring]-ysq)\
                -2*sqrt(self.rsquares[scattering_ring-1]-ysq)+abs(point[0]))
            j = scattering_ring-1
            while j > 0 and self.rsquares[j-1] > ysq:
                #summing up intermediate rings
                if ring_present[j]:
                    opath += 2*self.labs[j]*(sqrt(self.rsquares[j]-ysq)-\
                        sqrt(self.rsquares[j-1]-ysq))
                j += 1
            #Add contribution of innermost ring
            if ring_present[j]:
                opath += 2*self.labs[j]*sqrt(self.rsquares[j]-ysq)
        elif ring_present[scattering_ring]:
            #Central ring is the innermost ring touching the path
            opath += self.labs[scattering_ring]*sqrt(self.rsquares[scattering_ring]-ysq)
            opath_sc = self.labs[scattering_ring]*point[0]
            if incomingt_outgoingf:
                opath += opath_sc
            else:
                opath -= opath_sc
        return opath

    def __calc_single_ring_scattering(self, scattering_ring_no, layer_contributes, theta):
        "Returns with the absorption coefficient belonging to the scattering_ring_no, taking\
        into account layer_contributes layers as active absorbers and particle scattered to\
        2*theta degrees"
        sumvals = 0.0
        if scattering_ring_no is 0:
            no_base_rings = self.no_inner_rings[scattering_ring_no]
            start_index_i = 0
        else:
            no_base_rings = int(round(self.no_inner_rings[scattering_ring_no]/(1.0\
                -self.thickness[scattering_ring_no-1]/float(self.thickness[scattering_ring_no]))))
        rstep = self.thickness[scattering_ring_no]/float(no_base_rings)
        if scattering_ring_no > 0:
            start_index_i = no_base_rings - int(round((self.thickness[scattering_ring_no]\
                -self.thickness[scattering_ring_no-1])/rstep))
        for i in range(start_index_i, no_base_rings):
            if no_base_rings > 100 and ( i==start_index_i or i%100 == 0):
                print 'For theta='+str(theta)+' the ring no. are: '+str(i)+"/"+str(no_base_rings)
            ringsum = 0.0
            radius = (i+0.5)*rstep
            no_segments = self.start_segments[scattering_ring_no]*(i+1)
            phi = np.array([radians(theta)-pi/2.-0.5*pi/no_segments, 0.])
            phi[1] = phi[0]-2*radians(theta)
            for _ in range(no_segments):
                phi += pi/no_segments
                path = self.__calculate_optical_path(scattering_ring_no,\
                    np.array([radius*cos(phi[0]), radius*sin(phi[0])]),\
                            layer_contributes, True)
                path += self.__calculate_optical_path(scattering_ring_no,\
                    np.array([radius*cos(phi[1]), radius*sin(phi[1])]),\
                            layer_contributes, False)
                ringsum += exp(-path)
            sumvals += ringsum*(1.-0.5/(i+1))
        return 2.0*sumvals/(float(self.start_segments[scattering_ring_no]\
            *(no_base_rings**2-start_index_i**2)))

    def get_number_of_scatterers(self, beam_width, beam_height, layer_indices, theta_incoming=None):
        "Returns with the number of scatterers in mole units. Inputs are beam profile are in mm,\
        layer indices of the sample starting from zero from the direction of source and incoming\
        theta degree"
        area = 0.0
        for i in layer_indices:
            area += self.__circle_area(self.thickness[i], beam_width/20.0)*self.num_dens[i]
            if i > 0:
                area -= self.__circle_area(self.thickness[i-1], beam_width/10.0)*self.num_dens[i]
        return area*beam_height/10.0

    def calc_single_scattering(self, layer_contributes, layer_scattering, theta_incoming,\
        theta_outgoing=None):
        "Calculates the transmission coefficients belonging to layer_scattering layers, if\
        layer_contributes layers arepresent (these two are boolean arrays) at theta_incoming\
        and theta outgoing degrees. If outgoing degrees are not provided, symmetrical condition\
        is supposed"
        trsm = np.zeros(2)
        for i in range(len(self.labs)):
            if layer_scattering[i]:
                ssc = self.__calc_single_ring_scattering(i, layer_contributes, theta_incoming)
                #take surely larger than biggest diameter
                nsc = self.get_number_of_scatterers(21*self.thickness[-1], 1.0, np.array([i]))
                trsm[0] += ssc*nsc
                trsm[1] += nsc
        return trsm[0]/trsm[1]

class DoubleTripleLayerSymmetryicTThetaSecSample(object):
    "This class describes a double or triple layer |holder|sample (|holder| optional) geometry,\
    where angles are symmetrical and having the same thickness and material for the holder\
    before and after if triple geometry is used"
    def __init__(self, cls, sample_thickness_labs_numdensity, holder_thickness_labs_numdensity,\
        doublelayer=True, ringsegment=None):
        "Specifies the thickness of the sample and wall thickness of the holder in mm units and\
        the corresponding linear absorption coefficients in 1/cm units. Finally, the number\
        density of the sample and holdershould be provided in atom/A**3 units. The cls is the name\
        of the class describing the geometry, doublelayer distinguish between double or triple\
        layers. If (ring, segment, ring,  segment) is provided they are set accordingly, otherwise\
        determined them using empirical formulas"
        if cls is CylindricalGeometry:
            lin_abs = np.array([sample_thickness_labs_numdensity[1],\
                holder_thickness_labs_numdensity[1]], dtype=np.float)
            num_dens = np.array([sample_thickness_labs_numdensity[2],\
                holder_thickness_labs_numdensity[2]], dtype=np.float)
            lin_sizes = np.array([sample_thickness_labs_numdensity[0],\
                holder_thickness_labs_numdensity[0]], dtype=np.float)
            self.geometry = cls(lin_sizes, num_dens, lin_abs, ringsegment)
        else:
            lin_abs = np.array([holder_thickness_labs_numdensity[1],\
                sample_thickness_labs_numdensity[1]], dtype=np.float)
            num_dens = np.array([holder_thickness_labs_numdensity[2],\
                sample_thickness_labs_numdensity[2]], dtype=np.float)
            lin_sizes = np.array([holder_thickness_labs_numdensity[0],\
                sample_thickness_labs_numdensity[0]], dtype=np.float)
            if doublelayer is False:
                lin_abs = np.append(lin_abs, holder_thickness_labs_numdensity[1])
                num_dens = np.append(num_dens, holder_thickness_labs_numdensity[2])
                lin_sizes = np.append(lin_sizes, holder_thickness_labs_numdensity[0])
            self.geometry = cls(lin_sizes, num_dens, lin_abs)
        return

    def get_number_of_scatterers(self, beam_area, theta_incoming):
        "Returns with the number of scatterers in mole units. Inputs are beam profile are in cm^2,\
        incoming theta degree"
        return self.geometry.get_number_of_scatterers(beam_area, np.array([1]), theta_incoming)

    def get_transmission_coefficients(self, theta_incoming):
        "Calculates the transmission coefficients for Ts,s; Ts,sc; Tc,c; Tc,sc"
        full, holder = [np.ones(len(self.geometry.labs)) for _ in range(2)]
        sample = np.zeros(len(self.geometry.labs))
        if isinstance(self.geometry, FlatPlateGeometry):
            number = 1
        else:
            number = 0
        sample[number] = True
        holder[number] = False
        return self.geometry.calc_single_scattering(sample, sample, theta_incoming),\
            self.geometry.calc_single_scattering(full, sample, theta_incoming),\
                self.geometry.calc_single_scattering(holder, holder, theta_incoming),\
                    self.geometry.calc_single_scattering(full, holder, theta_incoming)

class NDToCorrected(object):
    "This class creates corrected dataset from normalized, deadtime corrected one"
    def __init__(self, filenames, sample_material, thicknesses_in_mm, holder_labs=0.,\
        polarization=0., cos_monottheta_sq=0., allow_extrapolation=False):
        "Initializes the dataset: creates basic structures and reads them. Inputs are filenames,\
        containing the datafiles of sample (and optionally: empty, background) order; Material\
        object of the sample; thicknesses of sample (then empty and background) in mm; linear\
        absorption coefficient of the holder; polarization parameter; cos^2(2theta_monochr),\
        the extrapolation switch (if it is false, onle the overlapping range of each data will be\
        corrected)"
        self.q_values = np.zeros(0)
        self.std_values = np.zeros(0) #standard format corresponds to q_values (2theta/wavelength)
        self.intensity = []
        self.dintensity = []
        self.standard = self.__read_and_standardize(filenames, allow_extrapolation)
        #standard data: constant wavelength or 2theta of the detector
        self.task_indicator = len(filenames)-1 #0: sample; 1: sample+empty; 2: sample+empty+bckg
        #Now, we are dealing with case of cylindrical geometry only
        sample = np.zeros(3)
        holder = np.zeros(3)
        sample[0] = thicknesses_in_mm[0]
        sample[1] = sample_material.get_lin_abs_coeff()
        sample[2] = sample_material.get_number_density()
        if len(filenames) > 1:
            holder[0] = thicknesses_in_mm[1]+thicknesses_in_mm[0]
        else:
            holder[0] = 0.0001
        holder[1] = holder_labs
        holder[2] = 0.1
        self.coherents = np.zeros_like(self.q_values)
        self.incoherents = np.zeros_like(self.q_values)
        self.inelastics = np.zeros_like(self.q_values)
        for i, q_val in enumerate(self.q_values):
            self.coherents[i] = sample_material.get_coherent(q_val)
            self.incoherents[i] = sample_material.get_incoherent(q_val)
            self.inelastics[i] = sample_material.get_inelastic(q_val)
        self.actual_polarization = 1.0 - polarization +\
            polarization * cos_monottheta_sq * np.cos(np.deg2rad(self.std_values))**2
        self.absorption_descriptor = DoubleTripleLayerSymmetryicTThetaSecSample(\
            CylindricalGeometry, sample, holder)
        return

    def __get_nd_dataset(self, filename):
        "Reads the normalized, deadtime corrected dataset in, returns with Q, I, dI and standard\
        on the calibrated range"
        data = np.loadtxt(filename, unpack=True)   #Q, I, dI, standard, native, calibrated
        mask = np.zeros(len(data[5]), dtype=bool)
        mask = (data[5] == 1.0)
        return data[0][mask], data[1][mask], data[2][mask], data[3][mask]

    def __read_and_standardize(self, filenames, allow_extrapolation=False):
        "Reads the normalized dataset from filenames array {order: sample[, empty[, background]]}\
        in, sweep them and assigns. Returns with the wavelength or 2Theta for angular and energy\
        dispersive sets, respectively. If solely empty or empty and background is provided program\
        will prepare them for the correction. In the case of difference between Q points of the\
        data, the sample data have thepriority. In the valid range, interpolation is allowed --\
        if extrapolation is not allowed, not suitable points will be omitted"
        q_orig = []
        i_orig = []
        di_orig = []
        std_orig = np.zeros(0)
        #read the files in
        for i, fil in enumerate(filenames):
            q_data, i_data, di_data, std_data = self.__get_nd_dataset(fil)
            q_orig.append(q_data)
            i_orig.append(i_data)
            di_orig.append(di_data)
            if i == 0:
                std_orig = std_data
        #Set Q values, trimming them if it were necessary -- do it for sample datasets, as well
        if allow_extrapolation:
            self.q_values = q_orig[0]
            self.std_values = std_orig
            self.intensity.append(i_orig[0])
            self.dintensity.append(di_orig[0])
        else:
            mask = q_orig[0][:]+.000001 >= max(q_orig, key=lambda x: x[0])[0]
            mask &= q_orig[0][:]-.000001 <= min(q_orig, key=lambda x: x[-1])[-1]
            self.q_values = q_orig[0][mask]
            self.std_values = std_orig[mask]
            self.intensity.append(i_orig[0][mask])
            self.dintensity.append(di_orig[0][mask])
        #interpolate and save other sets
        for i in range(1, len(filenames)):
            self.intensity.append(np.interp(self.q_values, q_orig[i], i_orig[i]))
            self.dintensity.append(np.interp(self.q_values, q_orig[i], di_orig[i]))
        #determine the applied wavelength/2theta angle
        if abs(sin(radians(std_orig[0]/2.0))/q_orig[0][0]-\
            sin(radians(std_orig[-1]/2.0))/q_orig[0][-1]) < 0.01:
            return 4*pi*sin(radians(std_orig[-1]/2.0))/q_orig[0][-1]
        else:
            print "ERROR: Energy dispersive setup is not implemented yet!"
            return 0.0

    def correct_emptyabspolarisation(self, outfile_name, write_all=False, abs_correction_step=5.):
        "Corrects the dataset by absorption, empty can and polarization, as well"
        #Geting the absorption correction coefficients
        start = int(self.std_values[0]/abs_correction_step) * abs_correction_step
        end = (int(self.std_values[-1]/abs_correction_step)+1) * abs_correction_step
        if end < self.std_values[-1]:
            end += abs_correction_step
        deg_points = np.arange(start, end+abs_correction_step, abs_correction_step)
        if deg_points[0] <= self.std_values[0]:
            deg_points[0] = self.std_values[0]
        else:
            np.insert(deg_points, 0, self.std_values[0])
        if deg_points[-1] >= self.std_values[-1]:
            deg_points[-1] = self.std_values[-1]
        else:
            np.append(deg_points, self.std_values[-1])
        trsmr_ss = np.zeros_like(deg_points)
        trsmr_ssc = np.zeros_like(deg_points)
        trsmr_cc = np.zeros_like(deg_points)
        trsmr_csc = np.zeros_like(deg_points)
        for i, degree in enumerate(deg_points):
            trsmr_ss[i], trsmr_ssc[i], trsmr_cc[i], trsmr_csc[i] = self.absorption_descriptor.\
                get_transmission_coefficients(degree/2.)
        trsm_ss = np.interp(self.std_values, deg_points, trsmr_ss)
        trsm_ssc = np.interp(self.std_values, deg_points, trsmr_ssc)
        trsm_cc = np.interp(self.std_values, deg_points, trsmr_cc)
        trsm_csc = np.interp(self.std_values, deg_points, trsmr_csc)
        corrected = np.zeros_like(self.q_values)
        dcorrected = np.zeros_like(self.q_values)
        if self.task_indicator == 0:
            corrected = self.intensity[0]/trsm_ss
            dcorrected = self.dintensity[0]/trsm_ss
        else:
            corrected = self.intensity[0] - trsm_csc/trsm_cc * self.intensity[1]
            dcorrected = self.dintensity[0]**2+(trsm_csc/trsm_cc* self.dintensity[1])**2
            if self.task_indicator == 2:
                corrected -= (1.0 - trsm_csc/trsm_cc) * self.intensity[2]
                dcorrected += ((1.0 - trsm_csc/trsm_cc) * self.dintensity[2])**2
            corrected /= trsm_ssc
            dcorrected = np.sqrt(dcorrected)/trsm_ssc
        corrected /= self.actual_polarization
        dcorrected /= self.actual_polarization
        #For special points, the transmission parameters are calculated,
        #calculate for the specific points
        hdr = 'Q, Corrected, dCorrected, Coherent, Incoherent, Inelastic, Standard, '
        dataset = [self.q_values] + [corrected] + [dcorrected] + [self.coherents] +\
            [self.incoherents] + [self.inelastics] + [self.std_values]
        if write_all:
            hdr += 'Polarization, Transmissions: (s,s; s,s+c; c,c; c,s+c), ' +\
                'Normalised intensity of sample'
            dataset += [self.actual_polarization] + [trsm_ss] + [trsm_ssc] + [trsm_cc] +\
                [trsm_csc] + [self.intensity[0]]
            if self.task_indicator >= 1:
                hdr += ', container'
                dataset += [self.intensity[1]]
            if self.task_indicator == 2:
                hdr += ', background'
                dataset += [self.intensity[2]]
        filen = outfile_name
        if outfile_name is not sys.stdout:
            filen += '_corr.dat'
        np.savetxt(filen, np.column_stack(dataset), header=hdr)

def help_dcorrection_controlfile():
    "Prints data correctin control file infos to screen"
    print "The control file should contain the following items:"
    print "1st line: #detectors #positons"
    print "2nd line: sample_file empty_file background_file" +\
        " -- the \"_N_ndt.dat\" extension will be automatically added"
    print "3rd line: N(eutron)/X(ray)"
    print "\twavelength[A]/energy[eV][only X: keV][only N: meV]/velocity[only N: m/s]"
    print "\t\t[for X: initial_polarisation monochromator_2theta_angle]"
    print "4th line: sample: length_in_mm fullness #components {if>0 densityN(umber)/G[/cm3]"
    print "\tthen each components in new line: symbol thencontribution}"
    print "{(if==0 and empty or background) linear_absorption_coefficient_in_cm^-1}"
    print "The same for empty and background, respectively"

def read_datacorrection_controlfile(filename):
    "This function reads the data correction control file. Returns with the radiation indicator,\
    initial polarization parameter and cos(2theta_monochromator), number of data, array of\
    filenames, the linear absorption coefficients, the linear lengths of each constituents, the\
    Material object related to the sample"
    with open(filename, 'r') as fil:
        #detectors and positions
        line = fil.readline().strip().split()
        no_data = int(line[0]) * int(line[1])
        #filenames: sample, empty, background
        filenames = fil.readline().strip().split()
        #radiation, wavelength/energy, polarization parameter
        line = fil.readline().strip().split()
        n_lx_sx = 0
        if line[0].upper() == "N":
            n_lx_sx = 0
            print "Source: neutron"
        elif line[0].upper() == "X":
            n_lx_sx = 1
            print "Source: x-ray"
        else:
            print "ERROR: unknown typeof radiation: \""+line[0].upper()+\
                "\" -- only N(eutron)/X(-ray) are allowed"
            exit()
        wlength = line[1]
        print "Monocromatic radiation parameter: ", wlength
        polarization = 0.0
        cos_monottheta_sq = 0.0
        if n_lx_sx > 0:
            polarization = float(line[2])
            cos_monottheta_sq = cos(radians(float(line[3])))**2
            print "Beam polarisation: " + str(polarization) + ", monochromator 2theta angle: "+\
                str(np.rad2deg(acos(sqrt(cos_monottheta_sq)))) + " degrees"
        lin_abs_coeffs = np.zeros(len(filenames))
        length_mm = np.zeros(len(filenames))
        #sample: length_in_mm fullness #components/ if zero: linear_absorption_coefficient,
            #if not:density; then symbol and contribution
        print "Parameters for ",
        for i, _ in enumerate(filenames):
            if i == 0:
                print "sample:"
            elif i == 1:
                print "empty:"
            elif i == 2:
                print "background:"
            line = fil.readline().strip().split()
            length_mm[i] = float(line[0])
            print "\tLinear size: "+str(length_mm[i])+" mm"
            fullness = float(line[1])
            print "\tFullness: "+str(100.0*fullness)+"%"
            ncomp = int(line[2])
            if ncomp == 0:
                if i == 0:
                    print "ERROR: you should provide the composition at least for sample!"
                    exit()
                else:
                    lin_abs_coeffs[i] = float(line[3])*fullness
            else:
                print "\tDensity: "+line[3]
                print "\tNumber of components: "+str(ncomp)
                material = Material(float(line[3][:-1]), fullness)
                for _ in range(ncomp):
                    linei = fil.readline().strip().split()
                    print "\t\tIsotope/element symbol: \""+linei[0]+"\", contribution: "+linei[1]
                    if n_lx_sx > 0:
                        material.add_element(XElement(linei[0], wlength), float(linei[1]))
                    else:
                        material.add_element(NElement(linei[0], wlength), float(linei[1]))
                if i == 0:
                    sample = material
                if (line[3][-1]).upper() == "N":
                    material.set_number_density(float(line[3][:-1]))
                elif line[3][-1].upper() != "G":
                    print "ERROR: the density should be marked by 1.123N as number density"+\
                        " or 1.123G as g/cm^3"
                    exit()
		else:
		    print "\tNumber density of the material is "+str(material.get_number_density())\
			+" 1/A^3"
		lin_abs_coeffs[i] = material.get_lin_abs_coeff()
	    print "\tLinear absorption coefficient: "+str(lin_abs_coeffs[i])+" 1/cm"
    return n_lx_sx, polarization, cos_monottheta_sq, no_data, filenames, lin_abs_coeffs,\
        length_mm, sample

def main(argv):
    "Reads the control datafile in"
    if len(argv) != 2:
        print "This program corrects sample dataset by background/empty container and absorption"
        print "\tUsage: "+argv[0]+" control_file.name"
        help_dcorrection_controlfile()
        exit()
    else:
        #Read the data correction file
        task_indicator, polar, c2mono, no_data, head_names, labs, lengths_in_mm, sample_material =\
            read_datacorrection_controlfile(argv[1])
        for i in range(no_data):
            filenames = list(head_names)
            outfile = head_names[0] + "_" + str(i+1)
            for j, _ in enumerate(filenames):
                filenames[j] += "_"+str(i+1)+"_ndt.dat"
            print filenames
            data = NDToCorrected(filenames, sample_material, lengths_in_mm, labs[1],\
                polar, c2mono, allow_extrapolation=False)
            data.correct_emptyabspolarisation(outfile, True)
    return
#
#   element3 = XElement(argv[1], argv[2])
#    print 'F(0): ', element3.get_f(0.0), ', Compton(0.): ', element3.get_compton(0.0)
#    sample = np.array([.2, 0.1, 0.1])
#    abscorr = DoubleTripleLayerSymmetryicTThetaSecSample(FlatPlateGeometryReflection,\
#      sample, holder)
#    print 'Flat plate reflection, double layer:', abscorr.get_transmission_coefficients(20.)
#    abscorr2 = DoubleTripleLayerSymmetryicTThetaSecSample(FlatPlateGeometryReflection,\
#      sample, holder, False)
#    print 'Flat plate reflection, triple layer:', abscorr2.get_transmission_coefficients(20.)
#    abscorr3 = DoubleTripleLayerSymmetryicTThetaSecSample(FlatPlateGeometryTransmission,\
#      sample, holder)
#    print 'Flat plate transmission, double layer:', abscorr3.get_transmission_coefficients(20.)
#    abscorr4 = DoubleTripleLayerSymmetryicTThetaSecSample(FlatPlateGeometryTransmission,\
#      sample, holder, False)
#    print 'Flat plate transmission, triple layer:', abscorr4.get_transmission_coefficients(20.)
#    holder = np.array([2.0, 3.0, 0.1])

#    mu_sample = float(argv[1])
#    mu_holder = float(argv[2])
#    print "deg, Ass, Assc, Acc, Acsc"
#    radius_in = float(argv[3])
#    radius_out = float(argv[4])+radius_in
#    ndensity_sample = float(argv[5])
#    ndensity_holder = float(argv[6])
#    sample = np.array([radius_in, mu_sample, ndensity_sample])
#    holder = np.array([radius_out, mu_holder, ndensity_holder])
#    abscorr5 = DoubleTripleLayerSymmetryicTThetaSecSample(CylindricalGeometry,\
#        sample, holder)
    #print 'Cylindrical shape, two rings:',
#    for degree in np.arange(181.):
#        print degree, abscorr5.get_transmission_coefficients(degree/2.)
#    print "Testing f calculation features..."
#    print "Tl: f(3.)=", element[0].get_f(3.), ' c(3.)=', element[0].get_compton(3.)
#    print "Ca: f(3.)=", element[1].get_f(3.), ' c(3.)=', element[1].get_compton(3.)
#    print "Ba: f(3.)=", element[2].get_f(3.), ' c(3.)=', element[2].get_compton(3.)
#    print "Cu: f(3.)=", element[3].get_f(3.), ' c(3.)=', element[3].get_compton(3.)
#    print "O: f(3.)=", element[4].get_f(3.), ' c(3.)=', element[4].get_compton(3.)
#    print '<f>=', mat.get_f_aver(3.), ' <f^2>=', mat.get_f2_aver(3.), '<f>^2=',\
#        mat.get_f_aver2(3.), ' Compton=', mat.get_compton(3.)

if __name__ == "__main__":
    main(sys.argv)
