#!/usr/bin/python
"The present module deals with the first part of data evaluation: normalization and\
deadtime correction"
from abc import ABCMeta, abstractmethod
from math import pi, cos, sin, radians
import os
import sys
import matplotlib.pyplot as plt
import numpy as np

def find_nearest(array, value):
    "Look for the nearest value in array to value. Returns with the nearest value and its index"
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx

class RawExperimentalDataSingleSegment(object):
    "This class deals with a single segment raw experimental dataset to produce standardized\
    output for further calculations. Summation and detector related corrections are also\
    performed here."
    __metaclass__ = ABCMeta
    objs = []           #instances
    n_st_labels = ["", ""]    #label for the native and standard value together with its unit

    def __init__(self, x_vals, native2standard_standard2q_params, native_standard_labels,\
        native_calibration_range, native_param2write=None, deadtime_usec=0.0, monitor_usec=0.0,\
            incrt_decrf=True):
        "Inputs are x_values (np array) in native units; parameter(s) necessary for\
        [ [native->standard(like channel->2Theta) parameters], [standard->Q]] transformation;\
        labels of the [native,standard] together its unit; calibrated range in native units\
        [min, max]; native parameter to write out in header, the deadtime of the detector then of\
        the monitor counter in microseconds; the increasing or decreasing order of native values"
        self.x_val = x_vals                         #native X coordinate values
        self.nat2std_std2q_params = native2standard_standard2q_params
                                                    #Parameters used to calculate
                                                    #[[standard from native], [Q from standard]]
        self.y_dy_norm = []                         #normalized, dt corrected datasets of each
                                                    # measurements with its corresponding errors
        self.y_dy_average = [np.zeros(len(self.x_val)), np.zeros(len(self.x_val))]
                                                    #average normalized dt corrected intensities,
                                                    # with its errors
        self.dt_usec = deadtime_usec                #deadtime in microsecs
        self.mon_usec = monitor_usec                #deadtime of the monitor counter in microsecs
        self.incrt_decf = incrt_decrf               #increase(t) or decrease (f) sorting of x values
        RawExperimentalDataSingleSegment.objs.append(self)
        self.n_st_labels = native_standard_labels   #Labels for [native, standard] together their
                                                    #units
        self.nativepar2write = native_param2write   #native parameter to write out in header
        if isinstance(native_calibration_range, np.ndarray):
            self.calibrated_min_max_x = native_calibration_range #Calibration range in native
                                                    #[min,max]
        else:
            self.calibrated_min_max_x = np.array(native_calibration_range)
        self.plot_figure = None                     #plotted figure subplot
        return

    def __del__(self):
        "Remove objects from objs list"
        print "Destructor"
        RawExperimentalDataSingleSegment.objs.remove(self)
        return

    def add_dataset(self, xyvals, monitor, detector_efficiency, measurement_seconds,\
        add_monitor_error=True):
        "Adds an experimental dataset to the structure. All inputs are numpy arrays. The xyvals\
        is an [N,N] element matrix contains the x and y subarrays. The size of other arrays is N,\
        correspondingly. Internally, deadtime correction and data normalization performed here.\
        If add_monitor_error is true, then counting error also taken into account"
        #Deadtime correction
        if np.any(measurement_seconds == 0.):
            print "WARNING: some of the measurement time is equal to zero.",\
                "At these values deadtime is not corrected"
        monmult = self._calc_dtmultiplicator(monitor, self.mon_usec, measurement_seconds)
        detmult = self._calc_dtmultiplicator(xyvals[1], self.dt_usec, measurement_seconds)
        dt_modulus = np.divide(monmult, detmult, out=np.ones_like(detmult), where=detmult != 0.)
        #Normalization
        ratio = np.divide(xyvals[1]/monitor/dt_modulus, detector_efficiency,\
            out=np.zeros_like(detector_efficiency),\
                where=detector_efficiency != 0.)
        errmult = np.divide(np.ones(len(xyvals[1]), dtype=float), xyvals[1],\
            out=np.zeros_like(xyvals[1]), where=xyvals[1] != 0.)
        if add_monitor_error:
            errmult += 1.0/monitor
        self._add_n_arrange_scale(xyvals[0], ratio, ratio*np.sqrt(errmult))
        self._calc_average()
        return

    def write_average(self, filename=sys.stdout, write_all=False):
        "Writes the Q; the normalized, deadtime corrected average data; the standard; the native\
        value; calibrated range indicator; optionally the individual sets into filename+'_ndt.dat'"
        std_val = self.transform_native2std()
        q_val = self.transform_std2q(std_val)
        hdr = 'Q, Norm, dNorm, '+self.n_st_labels[1]+', '+self.n_st_labels[0]+\
            ', calibrated(1-yes, 0-not)'
        dataset = [q_val] + self.y_dy_average + [std_val] + [self.x_val] +\
            [np.where((self.x_val >= self.calibrated_min_max_x[0]) &\
                (self.x_val <= self.calibrated_min_max_x[1]), 1, 0)]
        if write_all:
            hdr += ', '+str(len(self.y_dy_norm)/2)+\
                '*(Norm_i, dNorm_i [-1. means not present in the original dataset])'
            dataset += self.y_dy_norm
        if self.nativepar2write != None:
            hdr += ', other parameters: '+self.nativepar2write
        filen = filename
        if filename is not sys.stdout:
            filen += '_ndt.dat'
        np.savetxt(filen, np.column_stack(dataset), header=hdr)
        return

    def get_average(self):
        "This function returns with a list of columns containing Q, norm, dnorm, standard and\
        native values"
        std_val = self.transform_native2std()
        q_val = self.transform_std2q(std_val)
        dataset = [q_val] + self.y_dy_average + [std_val] + [self.x_val]
        return dataset

    def plot(self, fig):
        "Plots the average and all individual datasets to figure. The figure descriptor should be\
        provided"
        current_segment_no = 1+RawExperimentalDataSingleSegment.objs.index(self)
        max_segment_no = len(RawExperimentalDataSingleSegment.objs)
        cols = max_segment_no
        if max_segment_no > 5:
            cols = 5
        self.plot_figure = fig.add_subplot(1+(max_segment_no-1)/5, cols, current_segment_no)
        self.plot_figure.errorbar(self.x_val, self.y_dy_average[0], yerr=self.y_dy_average[1],\
            fmt='.')
        maxval = -10.
        for i in range(len(self.y_dy_norm)/2):
            self.plot_figure.plot(self.x_val, self.y_dy_norm[2*i])
            mval = np.amax(self.y_dy_norm[2*i]\
                [next(i for i, x in enumerate(self.x_val) if x >= self.calibrated_min_max_x[0]):\
                    next(i for i, x in enumerate(self.x_val) if x >= self.calibrated_min_max_x[1])])
            if maxval < mval:
                maxval = mval*1.1
        markers = self.calibrated_min_max_x
        for i in markers:
            self.plot_figure.axvline(x=i, color='r')
        lim, _ = self.plot_figure.get_ylim()
        if lim < 0.0:
            self.plot_figure.set_ylim(0.0,maxval)
        fig.canvas.mpl_connect('key_press_event', self._plot_switch_scale)
        self.plot_figure.set_xlabel(self.n_st_labels[0])
        self.plot_figure.set_title("Segment "+str(current_segment_no))
        if current_segment_no == 1:
            self.plot_figure.set_ylabel(r'Normalized by monitor counter')
            print "Press 'Q' for Q, 't' for "+self.n_st_labels[1]+" or 'n' to "+\
                self.n_st_labels[0]+"-scale view"
        return

    @staticmethod
    def _plot_switch_scale(event):
        "Switches the x-scale of the figure by pressing a button"
        sys.stdout.flush()
        if event.key == 'n' or event.key == 'Q' or event.key == 't':
            for obj in RawExperimentalDataSingleSegment.objs:
                obj.plot_figure.clear()
                if event.key == 'n':
                    #Switch to native-scale view
                    x_scale = obj.x_val
                    obj.plot_figure.set_xlabel(obj.n_st_labels[0])
                    markers = obj.calibrated_min_max_x
                elif event.key == 't':
                    #Switch to standard-scale view
                    x_scale = obj.transform_native2std()
                    obj.plot_figure.set_xlabel(obj.n_st_labels[1])
                    markers = obj.transform_native2std(obj.calibrated_min_max_x)
                elif event.key == 'Q':
                    #Switch to Q-scale view
                    std_scale = obj.transform_native2std()
                    x_scale = obj.transform_std2q(std_scale)
                    markers = obj.transform_std2q(\
                        obj.transform_native2std(obj.calibrated_min_max_x))
                    obj.plot_figure.set_xlabel(r'Q [$\AA^{-1}$]')
                obj.plot_figure.errorbar(x_scale, obj.y_dy_average[0], yerr=obj.y_dy_average[1],\
                    fmt='.')
                for i in range(len(obj.y_dy_norm)/2):
                    obj.plot_figure.plot(x_scale, obj.y_dy_norm[2*i])
                for i in markers:
                    obj.plot_figure.axvline(x=i, color='r')
                lim, _ = obj.plot_figure.get_ylim()
                if lim < 0.0:
                    obj.plot_figure.set_ylim(0.0,)
                current_segment_no = 1+RawExperimentalDataSingleSegment.objs.index(obj)
                obj.plot_figure.set_title("Segment "+str(current_segment_no))
                if current_segment_no == 1:
                    obj.plot_figure.set_ylabel(r'Normalized by monitor counter')
            plt.show()
        return

    def _calc_average(self):
        "Calculates the average y values taking into account each datasets"
        w_nominator = np.zeros(len(self.x_val))
        w_denominator = np.zeros(len(self.x_val))
        to_be_counted = []
        for i in range(len(self.y_dy_norm)/2):
            term = np.divide(np.ones(len(self.y_dy_norm[2*i+1]), dtype=float),\
                self.y_dy_norm[2*i+1]**2, out=np.zeros_like(self.y_dy_norm[2*i+1]),\
                    where=self.y_dy_norm[2*i+1] > 0.)
            #1.0/self.y_dy_norm[2*i+1]**2*(self.y_dy_norm[2*i+1] > -0.5)
            w_denominator += self.y_dy_norm[2*i]*term
            w_nominator += term
            to_be_counted.append(self.y_dy_norm[2*i+1] > -0.0)
        self.y_dy_average[0] = np.divide(w_denominator, w_nominator,\
            out=np.zeros_like(w_nominator), where=w_nominator > 0.)
        #w_denominator/w_nominator
        self.y_dy_average[1] = np.sqrt(np.divide(np.ones(len(w_nominator), dtype=float),\
            w_nominator, out=np.zeros_like(w_nominator), where=w_nominator > 0.))
        #1.0/np.sqrt(w_nominator)
        divisor = np.sum(to_be_counted, axis=0)
        divisor -= (divisor > 1)
        divisor = np.sqrt(divisor)
        self.y_dy_average[1] = np.divide(self.y_dy_average[1], divisor,\
            out=np.zeros_like(divisor), where=divisor > 0)
        return

    def _add_n_arrange_scale(self, x_scale, y_vals, dy_vals):
        "This method sweeps the already stored and the present dataset"
        if len(x_scale) == len(self.x_val):
            diff = np.max(np.fabs(x_scale - self.x_val))
            smstep = np.min(np.fabs(np.ediff1d(x_scale)))
            if diff < smstep*0.05:
                #Add to the present arrays
                self.y_dy_norm.append(y_vals)
                self.y_dy_norm.append(dy_vals)
                return
        print 'WARNING: the native x coords of the present dataset'+\
        ' does not fit with the previos one!'
        sweep_index, new_size = self._sweep_ndx(x_scale, )
        #sweep x coordinates
        old_x = self.x_val.copy()
        self.x_val = np.zeros(new_size)
        for i, j in enumerate(x_scale):
            self.x_val[sweep_index[1][i]] = j
        for i, j in enumerate(old_x):
            self.x_val[sweep_index[0][i]] = j
        #sweep y and dy
        for i, j in enumerate(self.y_dy_norm):
            new_pos = np.full(new_size, -1.)
            for k, old_xvals in enumerate(j):
                new_pos[sweep_index[0][k]] = old_xvals
            self.y_dy_norm[i] = new_pos
        new_pos = [np.full(new_size, -1.), np.full(new_size, -1.)]
        for i, j in enumerate(y_vals):
            new_pos[0][sweep_index[1][i]] = j
            new_pos[1][sweep_index[1][i]] = dy_vals[i]
        self.y_dy_norm.append(new_pos[0])
        self.y_dy_norm.append(new_pos[1])
        return

    def _sweep_ndx(self, new_x):
        "Sweeps the x values of the present and new_x coordinates organizing them in increasing\
        (t) or in decreasing (f) order according to previously set value"
        masks = [np.ones(len(self.x_val)), np.ones(len(new_x))]
        result = [np.zeros(len(self.x_val), dtype=int), np.zeros(len(new_x), dtype=int)]
        indx = 0
        sums = [masks[0].sum(), masks[1].sum()]
        while sums[0]+sums[1] > 0:
            if self.incrt_decf:
                if sums[0] != 0:
                    old_i = np.ma.array(self.x_val, mask=np.logical_not(masks[0])).argmin()
                if sums[1] != 0:
                    new_i = np.ma.array(new_x, mask=np.logical_not(masks[1])).argmin()
            else:
                if sums[0] != 0:
                    old_i = np.argmax(self.x_val*masks[0])
                if sums[1] != 0:
                    new_i = np.argmax(new_x*masks[1])
            if sums[0] == 0 or ((self.x_val[old_i] >= new_x[new_i] and self.incrt_decf) or\
                (self.x_val[old_i] <= new_x[new_i] and not self.incrt_decf))\
                    and sums[1] != 0:
                result[1][new_i] = indx
                masks[1][new_i] = False
            if sums[1] == 0 or ((self.x_val[old_i] <= new_x[new_i] and self.incrt_decf) or\
                (self.x_val[old_i] >= new_x[new_i] and not self.incrt_decf))\
                    and sums[0] != 0:
                result[0][old_i] = indx
                masks[0][old_i] = False
            sums = [masks[0].sum(), masks[1].sum()]
            indx += 1
        return result, indx

    @staticmethod
    def _calc_dtmultiplicator(measured_value, deadtime_usec, time_sec):
        "Calculates the multiplication factor of the measured value to reconstruct real intensity.\
        Inputs are the measured value, the deadtime in microseconds and measurement time in\
        seconds"
        convergence_crit = 1.e-8
        max_iteration = 100
        constant_term = np.divide(measured_value*deadtime_usec*1.0e-6, time_sec,\
            out=np.zeros_like(time_sec), where=time_sec != 0.)
        term_last = np.ones_like(constant_term)
        diff = term_last
        i = 0
        while abs(max(diff.min(), diff.max(), key=abs)) > convergence_crit and i < max_iteration:
            term = np.exp(term_last*constant_term)
            diff = np.divide(term-term_last, term_last, out=np.zeros_like(term_last),\
                where=term_last != 0.)
            #print i, term, term_last, diff
            term_last = term
            i += 1
        if i == max_iteration:
            print 'ERROR: in calc_dtmultiplicator the deadtime correction is not converged within'
            print str(max_iteration)+' iterations by the required accuracy! Only '+\
                str(abs(max(diff.min(), diff.max(), key=abs)))+' has been reached instead of '+\
                    str(convergence_crit)+'!'
            exit()
        return term

    @abstractmethod
    def transform_native2std(self, x_array=None):
        "Transforms x_array array in native units to standard units using nat2std_std2q_params"
        pass

    @abstractmethod
    def transform_std2q(self, x_array):
        "Transforms x_array array in standard units to Q-scale using nat2std_std2q_params"
        pass

class RawEnergyDispersiveSingleSegmentX(RawExperimentalDataSingleSegment):
    "This class deals with an x-ray energy dispersive one segment raw experimental dataset, to\
    produce standardized output for further calculations. Summation and detector related\
    corrections are also performed here. The standard format is energy array in keV, the\
    std2energy parameter is a single element numpy array containing the 2Theta angle of the\
    detector"
    __metaclass__ = ABCMeta
    def __init__(self, native_vals, ttheta_detector, native2energyparams, native_label,
                 native_calib_min, native_calib_max, native_param2write=None, deadtime_usec=0.0,\
                     monitor_usec=0.0):
        "Inputs are native (np array) x-values in arbitrary units; the ttheta_detector in degrees;\
        the native to E[keV] conversion parameters, the label of the native unit; calibration\
        range minimal; and maximal value in native units; native parameter to be appear in the\
        header; the deadtime of the detector and the monitor in microseconds"
        super(RawEnergyDispersiveSingleSegmentX, self).__init__(native_vals,\
            [native2energyparams, np.array(ttheta_detector)],\
                [native_label, 'E [keV]'], [native_calib_min, native_calib_max],\
                    native_param2write, deadtime_usec, monitor_usec, True)
        return

    def __del__(self):
        super(RawEnergyDispersiveSingleSegmentX, self).__del__()
        return

    def transform_std2q(self, x_array):
        "Transforms x_array array of E [keV] to Q [AA^-1]"
        return 1.01354620746*x_array*np.sin(np.radians(self.nat2std_std2q_params[1]/2.))

    @abstractmethod
    def transform_native2std(self, x_array=None):
        "Transforms x_array array in native units to standard units using nat2std_std2q_params"
        pass

class RawAngularDispersiveSingleSegment(RawExperimentalDataSingleSegment):
    "This class deals with an angular dispersive one segment raw experimental dataset to produce\
    standardized output for further calculations. Summation and detector related corrections are\
    also performed here. The standard format is 2Theta array in degrees, the std2ttheta parameter\
    is a single length numpy array containing the wavelength in Angstrom units"
    __metaclass__ = ABCMeta
    def __init__(self, native_vals, wavelength, native2tthetaparams, native_label,
                 native_calib_min, native_calib_max, native_param2write=None, deadtime_usec=0.0,\
                     monitor_usec=0.0):
        "Inputs are native (np array) x-values in arbitrary units; the wavelength in AA units;\
        the native to ttheta conversion parameters, the label of the native unit; calibration\
        range minimal; and maximal value in native units; native parameter to be appear in the\
        header; the deadtime of the detector and the monitor in microseconds"
        super(RawAngularDispersiveSingleSegment, self).__init__(native_vals,\
            [native2tthetaparams, np.array(wavelength)],\
                [native_label, '2Theta [degrees]'], [native_calib_min, native_calib_max],\
                    native_param2write, deadtime_usec, monitor_usec, True)
        return

    def __del__(self):
        print "290"
        super(RawAngularDispersiveSingleSegment, self).__del__()
        return

    def transform_std2q(self, x_array):
        "Transforms x_array array of 2Theta [degrees] to Q [AA^-1]"
        return 4.*pi/self.nat2std_std2q_params[1]*np.sin(np.radians(x_array/2.))

    @abstractmethod
    def transform_native2std(self, x_array=None):
        "Transforms x_array array in native units to standard units using nat2std_std2q_params"
        pass

class RawAngularDSingleSegmentTThetaDetector(RawAngularDispersiveSingleSegment):
    "This class describes an angular dispersive detector having 2Theta as native format.\
    It is assumed that it is fully calibrated over the whole range"
    def __init__(self, ttheta_vals, wavelength, deadtime_usec=0.0, monitor_usec=0.0):
        "Inputs are 2Theta values in numpy array, the wavelength in AA units and the deadtime of\
        the detector and monitor counter in microseconds"
        super(RawAngularDSingleSegmentTThetaDetector, self).__init__(ttheta_vals,\
            wavelength, 1, '2Theta [degrees]', ttheta_vals[ttheta_vals.argmin()],\
                ttheta_vals[ttheta_vals.argmax()], None, deadtime_usec, monitor_usec)
        return

    def __del__(self):
        super(RawAngularDSingleSegmentTThetaDetector, self).__del__()
        return

    def add_dataset(self, xyvals, monitor, detector_efficiency, measurement_seconds,\
        add_monitor_error=True):
        "Adds an experimental dataset to the structure. All (except detector effictor_efficiency)\
        inputs are numpy arrays. The xyvals is an [N,N] element matrix contains the ttheta and y\
        subarrays. The size of other arrays is N, correspondingly. Internally, deadtime correction\
        and data normalization performed here. It also updates the calibrated range to be full, if\
        it were necessary. Also, by default it adds the monitor counter error to data point error"
        super(RawAngularDSingleSegmentTThetaDetector, self).add_dataset(xyvals, monitor,\
            np.full_like(xyvals[0], detector_efficiency), measurement_seconds, add_monitor_error)
        #update calibration range to be full
        if self.x_val[self.x_val.argmin()] < self.calibrated_min_max_x[0]:
            self.calibrated_min_max_x[0] = xyvals[0][self.x_val.argmin()]
        if self.x_val[self.x_val.argmax()] > self.calibrated_min_max_x[1]:
            self.calibrated_min_max_x[1] = xyvals[0][self.x_val.argmax()]
        return

    def transform_native2std(self, x_array=None):
        "Transforms x_array 2theta array to the same unit so native=standard.\
        If x_array is not provided native x values are supposed"
        if x_array is None:
            x_array = self.x_val
        return x_array

class RawAngularPSDDetectorSingleSegment(RawAngularDispersiveSingleSegment):
    "This class describes an angular dispersive position sensitive detector at one segment"
    def __init__(self, channel_vals, wavelength, ch2ttheta_poly_coeffs,\
        reference_angle_ttheta, calib_channel_min, calib_channel_max, detector_efficiency,\
            rangle_segment_no=None, detector_name="", deadtime_usec=0.0, monitor_usec=0.0):
        "Inputs are the channel numbers; the wavelength in AA units; the channel to angle\
        conversion polynomials as numpy polynomials; the reference 2Theta angle; the calibration\
        range minimal; and maximal channel values; detector efficiency for each channels;\
        the reference angle and segment number in string; the detector name, the deadtime of the\
        detector in microseconds"
        super(RawAngularPSDDetectorSingleSegment, self).__init__(channel_vals,\
            wavelength, [reference_angle_ttheta, ch2ttheta_poly_coeffs], 'Channel',\
                calib_channel_min, calib_channel_max, rangle_segment_no, deadtime_usec,\
                    monitor_usec)
        self.efficiency4each_channels = detector_efficiency
        self.det_name = detector_name
        return

    def __del__(self):
        print "356"
        super(RawAngularPSDDetectorSingleSegment, self).__del__()
        return

    def transform_native2std(self, x_array=None):
        "Transforms the x_array array in channels to 2theta units. If x_array is not provided the\
        x_val array will be transformed "
        number_of_channels = len(self.x_val)
        polys = self.nat2std_std2q_params[0][1]
        if x_array is None:
            x_array = self.x_val
        diff = x_array*2./number_of_channels - 1.
        return self.get_reference_angle()+np.polyval(polys, diff)

    def add_dataset(self, xyvals, monitor, detector_efficiency, measurement_seconds,\
        add_monitor_error=True):
        raise NotImplementedError(\
            "This method is not available for objects of RawAngularPSDDetectorSingleSegment\n"+\
                "Please use the " + "add_data(self, xyvals, monitor, detector_efficiency,"+\
                    " measurement_seconds) instead of it")

    def add_data(self, yvals, monitor, measurement_seconds):
        "Adds an experimental dataset to the structure. The yvals is an N element numpy array\
        contains the y values; monitor is the monitor counts during measurement_seconds time.\
        Internally, deadtime correction and data normalization performed here."
        super(RawAngularPSDDetectorSingleSegment, self).add_dataset(\
            np.array([self.x_val, yvals]), np.full_like(self.x_val, monitor),\
                self.efficiency4each_channels, np.full_like(self.x_val, measurement_seconds),\
                    True)
        return

    def get_reference_angle(self):
        "Returns with the reference angle of the current segment"
        return self.nat2std_std2q_params[0][0]

    def get_detector_name(self):
        "Returns with the detector no."
        return self.det_name

class RawXEnergyDDetectorAtGivenAngle(RawEnergyDispersiveSingleSegmentX):
    "This class describes an energy dispersive detector at one angular position"
    def __init__(self, channel_vals, ttheta_angle, ch2energy_poly_coeffs, channel_efficienty,\
        calib_channel_min, calib_channel_max, deadtime_usec=0.0, monitor_usec=0.0):
        "Inputs are the channel numbers; the 2Theta in degrees; the channel to energy\
        conversion polynomials as numpy polynomials; the channel efficiencies as np array; the\
        calibration range minimal; and maximal channel values; the deadtime of the detector and\
        monitor counter in microseconds "
        super(RawXEnergyDDetectorAtGivenAngle, self).__init__(channel_vals,\
            ttheta_angle, ch2energy_poly_coeffs, 'Channel',\
                calib_channel_min, calib_channel_max, str(ttheta_angle), deadtime_usec,\
                    monitor_usec)
        self.efficiency4each_channels = channel_efficienty
        return

    def __del__(self):
        super(RawXEnergyDDetectorAtGivenAngle, self).__del__()
        return

    def transform_native2std(self, x_array=None):
        "Transforms the x_array array in channels to energy in keV. If x_array is not provided the\
        x_val array will be transformed "
        polys = self.nat2std_std2q_params[0]
        if x_array is None:
            x_array = self.x_val
        return np.polyval(polys, x_array)

    def add_dataset(self, xyvals, monitor, detector_efficiency, measurement_seconds,\
        add_monitor_error=True):
        raise NotImplementedError(\
            "This method is not available for objects of RawXEnergyDDetectorAtGivenAngle\n"+\
                "Please use the " + "add_data(self, yvals, measurement_seconds) instead of it")

    def add_data(self, yvals, measurement_seconds):
        "Adds an experimental dataset to the structure. The yvals is an N element numpy array\
        contains the y values; the non-deadtime corrected measurement in measurement_seconds time.\
        Internally, deadtime correction and data normalization performed here."
        super(RawXEnergyDDetectorAtGivenAngle, self).add_dataset(\
            np.array([self.x_val, yvals]), np.full_like(self.x_val, measurement_seconds),\
                self.efficiency4each_channels, np.full_like(self.x_val, measurement_seconds),\
                    False)
        return

    def get_tthetadeg(self):
        "Returns with the 2Theta angle of the current segment in degrees"
        return self.nat2std_std2q_params[1][0]


class RawAngularPSDDetectorSingleSegmentSystem(RawAngularPSDDetectorSingleSegment):
    "This class describes an angular dispersive position sensitive detector system at one segment\
    -- the present version uses the modified calibration system"
    def __init__(self, channel_vals, wavelength, vertical_position, shortest_sample_tube_distance,\
        tube_centre_perpendicular_distance, ch2tubelength_poly_coeffs, reference_channel,\
            reference_angle_ttheta, reference_angle_shift, lims, detector_efficiency,\
                detector_name, detector_no=1, deadtime_usec=0.0, monitor_usec=0.0):
        "Inputs are the channel numbers; the wavelength in AA units; the vertical position of the\
        tube in mm units; shortest sample-tube distance in mm; centre of tube reference from\
        perpendicular position in mm; the channel to angle conversion polynomials as numpy\
        polynomials; the reference channel, the reference 2Theta angle; the shift angle belonging\
        to the perpendicular position relative to the reference angle; limit vamues: {min_channel,\
        [optionally barriers], max_channel}; detector efficiency for each channels; name of the\
        detector, number of the detector, the deadtime of the detector and monitor counter in\
        microseconds"
        super(RawAngularPSDDetectorSingleSegmentSystem, self).__init__(channel_vals,\
            wavelength, np.divide(ch2tubelength_poly_coeffs[0],\
                shortest_sample_tube_distance), reference_angle_ttheta+\
                    reference_angle_shift, lims[0], lims[-1], detector_efficiency,\
                        str(reference_angle_ttheta)+' '+str(detector_no)+' '+detector_name,\
                            detector_name, deadtime_usec, monitor_usec)
        self.vertical_ratio = vertical_position/shortest_sample_tube_distance  #Ratio of the
            #vertical shift relative to the scattering plane and the tube-sample distance
        self.center_tangent = tube_centre_perpendicular_distance/shortest_sample_tube_distance
            #Tangent of the tube centre - shortest tube-sample point on the tube- sample in hplane
        if len(lims) > 2:
            self.barriers = lims[1:] #Barrier used to separate different calibration regimes
            self.coeffs_upper = ch2tubelength_poly_coeffs[1:]
            for i, j in enumerate(self.coeffs_upper):
                self.coeffs_upper[i] = np.divide(j, shortest_sample_tube_distance)
        else:
            self.barriers = None
            self.coeffs_upper = None
        self.ref_chn = reference_channel #reference channel for the dataset
        self.refa_shift = reference_angle_shift #reference angle shift
        return

    def __del__(self):
        print "435"
        super(RawAngularPSDDetectorSingleSegmentSystem, self).__del__()
        return

    def transform_native2std(self, x_array=None):
        "Transforms the x_array array in channels to 2theta units. If x_array is not provided the\
        x_val array will be transformed "
        if x_array is None:
            x_array = self.x_val
        r_array = x_array - self.ref_chn
        r_t = super(RawAngularPSDDetectorSingleSegmentSystem, self).get_reference_angle()
        poly = np.zeros(len(self.nat2std_std2q_params[0][1])+1)
        poly[1:] = self.nat2std_std2q_params[0][1]
        ang = np.add(np.polyval(poly[::-1], r_array), self.center_tangent)
        #Dealing with splitted range
        if self.barriers is not None:
            for i in range(len(self.barriers)-1):
                poly[1:] = self.coeffs_upper[i]
                up_val = int(self.barriers[i+1])
                if i == len(self.barriers)-2:
                    up_val = len(ang)
                ang[int(self.barriers[i])-1:up_val] = np.add(np.polyval(poly[::-1],\
                r_array[int(self.barriers[i])-1:up_val]), self.center_tangent)
        denominator = cos(radians(r_t)) - sin(radians(r_t)) * ang
        nominator = np.sqrt(1. + np.power(ang, 2) + self.vertical_ratio**2)
        #print np.degrees(np.arccos(denominator/nominator))
        #exit()
        return np.degrees(np.arccos(denominator/nominator))

    def get_reference_angle(self, non_zero_corrected=False):
        "Returns with the reference angle of the current segment"
        value = super(RawAngularPSDDetectorSingleSegmentSystem, self).get_reference_angle()
        if non_zero_corrected:
            value -= self.refa_shift
        return value

class MTESTPSDData(object):
    "This class transfroms raw experimental data into summarized, normalized format recorded at\
    MTEST diffractometer using single PSD setup"
    def __init__(self, wavelength, scanning_variable_label, filenames, lims,\
        polys, vertical_position, sample_tube_plane_distance, tube_distance_shift,\
            reference_angle_shifts, reference_channels, detector_efficiencies, deadtimes,\
                tolerance_in_reference_angle):
        "Initializes the class by definying the applied wavelength in angstroms; the label of\
        scanning motor; all of the filenames containing data to add; then for each individual PSD\
        detectors: limits where calibration is valid; [polys]_each_tubes: the channel->mm\
        transformation polys; vertical_shifts, sample-tube distances, tube_perpendicular_shift in\
        mm, reference angle shifts, reference_channels, detector efficienties for\
        [[each channel],..]; deadtimes in microseconds(first is the monitor, then regular\
        detectors);finally tolerance in terms of reference angle"
        self.segments = []
        for _, fil in enumerate(filenames):
            #get channel intensity data
            dataset = np.loadtxt(fil, unpack=True)
            reference_angle = None
            monitor = None
            no_data = []
            names = []
            with open(fil, 'r') as filp:
                for line in filp:
                    indx = line.find(scanning_variable_label)
                    indx2 = line.find("Number of counters:")
                    if indx != -1:
                        #get the reference angle
                        reference_angle = float(line.split()[-1])
                    if indx2 != -1:
                        #get the number of counters
                        no_counters = int(line.split()[-1])
                        filp.next() #skip the comment line
                        for i in range(no_counters):
                            line = filp.next()
                            splitted = line.split()
                            if line.find("Monitor") != -1:
                                monitor = int(splitted[4])
                                measurement_seconds = int(splitted[5])
                                no_counters -= 1
                            elif line.find("_amplitude") != -1:
                                no_counters -= 1
                            else:
                                no_data.append(int(splitted[2]))
                                names.append(splitted[1])
                        break
            if reference_angle is None or monitor is None:
                raise Exception('Nor reference angle, nor monitor counter is assigned!'+\
                    ' Please check the "'+fil+'" file!')
            #We should look for existing reference angles
            found_indx = -1
            for i, j in enumerate(self.segments):
                r_ang = j.get_reference_angle(True)
                if abs(r_ang-reference_angle) < tolerance_in_reference_angle and names[0] ==\
                    j.get_detector_name():
                    found_indx = i
                    break
            if found_indx == -1:
                #There is not any existing segment: create a new
                found_indx = len(self.segments)
                if len(detector_efficiencies) <= len(self.segments):
                    effstart = 0
                else:
                    effstart = found_indx
                for i, act in enumerate(no_data):
                    self.segments.append(\
                        RawAngularPSDDetectorSingleSegmentSystem(dataset[0][0:act], wavelength,\
                            vertical_position[i], sample_tube_plane_distance[i],\
                                tube_distance_shift[i], polys[i], reference_channels[i],\
                                    reference_angle, reference_angle_shifts[i], lims[i],\
                                        detector_efficiencies[effstart+i], names[i], i+1,\
                                            deadtimes[i+1], deadtimes[0]))
            #Add the dataset to the structure
            for i, act in enumerate(no_data):
                self.segments[found_indx+i].add_data(dataset[i+1][0:act],\
                    monitor, measurement_seconds)
        return

    def __del__(self):
        for _, i in enumerate(self.segments):
            del i
        del self.segments
        return

    def write_summarized2file(self, filename):
        "Writes the normalized, deadtime corrected files into filename+'_X_ndt.dat' file"
        for i, j in enumerate(self.segments):
            fname = filename
            if len(self.segments) > 1:
                fname += '_'+str(i+1)
            j.write_average(fname, True)
        return

    def write_summarized(self):
        "Writes the normalized, deadtime corrected files to stdout"
        for i, j in enumerate(self.segments):
            if len(self.segments) > 1:
                print '#Datasets for segment number ', i+1
            j.write_average()
        return

    def plot(self):
        "Plots the data to screen"
        fig = plt.figure()
        for _, j in enumerate(self.segments):
            j.plot(fig)
        plt.show()

    def get_no_segments(self):
        "Returns with the number of segments"
        return len(self.segments)

    def get_reference_angle(self, nth=0):
        "Returns the reference angle of the nth segment"
        return self.segments[nth].get_reference_angle()

class BL04B2Data(object):
    "This class transfroms raw experimental data into summarized, normalized format recorded at\
    BL04B2"
    def __init__(self, filename, angle_shifts, deadtime_usec, wavelength):
        "Initializes the structure and reads the datasets in. Inputs are the array of filenames,\
        the angle_shifts array, the deadtime in microseconds for each detectors (first is the\
        monitor deadtime) and the wavelength in Angstroms"
        #For new datasets: each segments are summarized here
        #For old datasets: only one segment is summarized here
        #Reading the dataset in:2 rows comment, 0=ttheta, 1=theta, 2=Q, 3=t, 4=monitor, 4+ =Counts
        if len(angle_shifts) != len(deadtime_usec)-1:
            raise Exception('BL04B2Data.__init__ The size of angle_shifts (', len(angle_shifts),\
                ') and deadtime_usec-1 (', len(deadtime_usec)-1,\
                    ') do not conform with each other')
        first = True
        self.segments = []
        self.angle_shifts = angle_shifts
        for _, j in enumerate(filename):
            dataset = np.loadtxt(j, unpack=True)
            if first:
                self.ttheta = dataset[0]
            for i in range(len(self.angle_shifts)):
                if first:
                    self.segments.append(\
                        RawAngularDSingleSegmentTThetaDetector(dataset[0]+self.angle_shifts[i],\
                            wavelength, deadtime_usec[i+1], deadtime_usec[0]))
                #Add the dataset to the structure
                self.segments[i].add_dataset([dataset[0]+self.angle_shifts[i], dataset[5+i]],\
                    dataset[4], 1., dataset[3], True)
            first = False
        return

    def write_summarized2file(self, filename):
        "Writes the normalized, deadtime corrected files into filename+'_X_ndt.dat' file"
        for i, j in enumerate(self.segments):
            fname = filename
            if len(self.segments) > 1:
                fname += '_'+str(i+1)
            j.write_average(fname, True)
        return

    def write_summarized(self):
        "Writes the normalized, deadtime corrected files to stdout"
        for i, j in enumerate(self.segments):
            if len(self.segments) > 1:
                print '#Datasets for segment number ', i+1
            j.write_average()
        return

    def plot(self):
        "Plots the data to screen"
        fig = plt.figure()
        for _, j in enumerate(self.segments):
            j.plot(fig)
        plt.show()

class BL04B1Data(object):
    "This class transfroms raw experimental data into summarized, normalized format recorded at\
    BL04B1"
    def __init__(self, filenames, calibrated_ttheta=None, efficiency=None):
        "Initializes the structure and reads the datasets in. Input is the filename array, the\
        2Theta calibration pair-arrays and efficiency for each channels"
        first = True
        for _, filen in enumerate(filenames):
            chk, chepoly, ttheta, real_time, no_header, no_datapts = self.__get_npd_header(filen)
            if chk:
                print 'ERROR(BL04B1Data::__init: the "', filen,\
                    '" file does not exist or not .NPD formatted!'
            else:
                if calibrated_ttheta is not None:
                    nval, nidx = find_nearest(calibrated_ttheta[0], ttheta)
                    newval = calibrated_ttheta[1][nidx]+ttheta-nval
                    print "The "+str(ttheta)+" 2Theta value is adjusted to "+str(newval)+\
                        " according to 2Theta calibration file"
                    ttheta = newval
                channels, counts = np.loadtxt(filen, delimiter=",", skiprows=no_header,\
                    usecols=(0, 2), unpack=True)
                if no_datapts != len(counts):
                    print 'ERROR(BL04B1Data::__init): mismatched header: number of available',\
                        ' data points (', len(counts), ') differ from that in the header (',\
                            no_datapts, ')!'
                else:
                    if first:
                        chepoly_ref = chepoly
                        ttheta_ref = ttheta
                        chn_ref = len(counts)
                        channel_efficienty = np.ones_like(counts)
                        if efficiency is not None:
                            if len(efficiency) != len(counts):
                                print 'ERROR(BL04B1Data::__init): the number of efficiency points',\
                                    ' (', len(efficiency), ') does not match with the available',\
                                        ' number of points (', len(no_datapts), ')!'
                                return
                            else:
                                channel_efficienty = efficiency
                        self.data = RawXEnergyDDetectorAtGivenAngle(channels, ttheta, chepoly,\
                            channel_efficienty, channels[0], channels[-1])
                        first = False
                    elif any(chepoly_ref != chepoly) or ttheta_ref != ttheta or len(channels) != chn_ref:
                        print 'ERROR(BL04B1Data::__init): mismatch between the available datasets:'
                        print '+channel->energy polynomial: ', chepoly_ref, ' vs ', chepoly
                        print '+2theta value: ', ttheta_ref, ' vs ', ttheta
                        print '+#channels: ', chn_ref, ' vs ', len(counts)
                        return
                    self.data.add_data(counts, real_time)
        return

    def __get_npd_header(self, filename):
        "Open the filename and returns with 1(success)/0(fail). In the case of success, returns\
        with the extracted channel->energy polynomial coeffs, 2theta, real time, number of header\
        lines, number of data points"
        good_format = False
        header_length = 0
        no_data_pts = 0
        ttdeg = 0.
        cnt_sec = 0.
        poly = np.zeros(5)
        if os.path.isfile(filename):
            with open(filename, 'r') as fil:
                counter = -1
                for line in fil:
                    if counter >= header_length:
                        break
                    else:
                        counter += 1
                        lin = line.strip().split(",")
                        if lin[0] == "Record length":
                            header_length = int(lin[1])+2
                        elif lin[0] == "2Theta(deg)":
                            ttdeg = float(lin[1])
                        elif lin[0] == "Data length":
                            no_data_pts = int(lin[1])
                        elif lin[0] == "Counting sec":
                            cnt_sec = float(lin[1])
                        elif lin[0].find("EGC") != -1:
                            poly[int(lin[0][3:])] = float(lin[1])
        return good_format, poly[::-1], ttdeg, cnt_sec, header_length, no_data_pts

    def write_summarized2file(self, filename):
        "Writes the normalized, deadtime corrected files into filename+'_ndt.dat' file"
        self.data.write_average(filename, True)
        return

    def write_summarized(self):
        "Writes the normalized, deadtime corrected files to stdout"
        self.data.write_average()
        return

    def plot(self):
        "Plots the data to screen"
        fig = plt.figure()
        self.data.plot(fig)
        plt.show()

