#!/usr/bin/python
"This program removes Bragg-peaks from vanadium datasets measured by powder diffraction with\
constant wavelength technique supposing Gaussian peak profiles"
from collections import OrderedDict
import sys
import numpy as np
import numpy.polynomial.polynomial as polyn
from scipy.optimize import minimize
from scipy.signal import savgol_filter as savgol

def determine_cubic_multiplicity(reflection):
    "Determines the multiplicity of a cubic lattice reflection recorded at powder diffraction.\
    Reflection is a numpy array containing integer indices"
    multiplicity = 1
    if reflection[0] == reflection[1] and reflection[1] == reflection[2]:
        multiplicity = 1
    elif reflection[0] != reflection[1] and reflection[1] != reflection[2] and\
        reflection[2] != reflection[0]:
        multiplicity = 6
    else:
        multiplicity = 3
    for _, ref in enumerate(reflection):
        if ref != 0:
            multiplicity *= 2
    return multiplicity

def calpattoneref_sig(degrees, central_degree, sigma):
    "Calculates the contribution of a Bragg-peak of intensity=1 for degrees points,\
    centered at central_degree, with sigma parameter"
    multiplicator = 1.0/np.sqrt(2 * np.pi)/sigma
    return multiplicator*np.exp(-0.5*((degrees-central_degree)/sigma)**2)

def calpattoneref_uvw(degrees, central_degree, resolution):
    "Calculates the contribution of a Bragg-peak of intensity=1 for degrees points,\
    centered at central_degree, with resolution parameters of U,V,W"
    sigma = np.sqrt((resolution[0]*(np.tan(np.deg2rad(central_degree)/2.))**2+\
        resolution[1]*np.tan(np.deg2rad(central_degree)/2.)+resolution[2])/8./np.log(2.))
    return calpattoneref_sig(degrees, central_degree, sigma)

def angleintensity_generator_bcc(wavelength, lattice_constant, ttheta_max, theta_min=0.):
    "Generates an ordered list of 2theta, multiplicity pairs, for bcc structures having\
    lattice_constant in A, measured with wavelength A up to ttheta_max degrees in 2theta"
    multiplicator = 2*lattice_constant/wavelength
    max_square = (multiplicator * np.sin(np.deg2rad(ttheta_max/2.)))**2
    bragg_peaks = OrderedDict()
    for indx_h in range(int(np.sqrt(max_square))+1):
        for indx_k in range(indx_h+1):
            hk_square = indx_h**2 + indx_k**2
            for indx_l in range(indx_k+1):
                if (indx_h+indx_k+indx_l)%2 == 0 and hk_square + indx_l**2 <= max_square:
                    degree = round(2*np.rad2deg(np.arcsin(np.sqrt(hk_square + indx_l**2)/\
                        multiplicator)), 7)
                    multiplicity = determine_cubic_multiplicity(np.array([indx_h, indx_k,\
                        indx_l]))
                    #Multiply by 2 ( [1+1]**2/#atoms_in_cell: 2/atoms)
                    if degree > theta_min:
                        if degree in bragg_peaks:
                            bragg_peaks[degree] += 2*multiplicity
                        else:
                            bragg_peaks[degree] = 2*multiplicity
    if theta_min == 0.:
        del bragg_peaks[0.0]    #delete the 000 reflection
    b_peaks_ordered = OrderedDict(sorted(bragg_peaks.items()))
    return b_peaks_ordered

def create_bragg_pattern(degrees, resolution, wavelength, lattice_constant, msqd):
    "Creates intensity of Bragg-pattern at degrees, if resolution (np array U,V,W), wavelength in\
    A, lattice_constant in A and msqd in A**2 are provided"
    peak_list = angleintensity_generator_bcc(wavelength, lattice_constant,\
        min(np.amax(degrees)+20., 180.))
    intensity = np.zeros_like(degrees)
    for _, item in enumerate(peak_list):
        q_value = 4.*np.pi/wavelength*np.sin(np.deg2rad(item/2.))
        intensity += calpattoneref_uvw(degrees, item, resolution)*\
            peak_list.get(item)*np.exp(-msqd*q_value**2/3.0)
    return intensity

def bckg_intensities_fwhm_pattern(parameters, x_data, y_data, bck_degree, central_angles,\
    wlength=0, uvw_pars=None):
    "This function returns with calculated pattern and the background contribution. If\
    bck_degree>0, then polynomial coeffs are read from parameters, otherwise Savitzky-Golay\
    filter going to use with -bck_degree polynomial degree and wlength window. Parameters then\
    contain intensity of individual Bragg-peaks and optionally UVW parameters based on x_data.\
    If uvw_pars is not empty, it performs the fitting task without fitting the Cagliotti parameters"
    #decoding parameters
    start = 0
    last = 0
    if bck_degree >= 0:
        last = bck_degree+1
        bckg = polyn.Polynomial(parameters[start:last])  # background coefficients
    # Bragg_intensities
    start = last
    last += len(central_angles)
    bragg_intensity = parameters[start:last]
    uvw = np.zeros(3)
    if uvw_pars is None:
        start = last
        last += 3
        uvw = parameters[start:last]
    else:
        uvw = uvw_pars
    #generate calculated dataset
    if bck_degree >= 0:
        y_calc = bckg(x_data)
        y_bckg = bckg(x_data)
    else:
        y_calc = np.zeros_like(x_data)
    for i, item in enumerate(central_angles):
        y_calc += abs(bragg_intensity[i])*calpattoneref_uvw(x_data, item, uvw)
    if bck_degree < 0:
        y_bckg = savgol(y_data-y_calc, wlength, -bck_degree)
        y_calc = y_bckg.copy()
        for i, item in enumerate(central_angles):
            y_calc += abs(bragg_intensity[i])*calpattoneref_uvw(x_data, item, uvw)
    return y_calc, y_bckg

def fit_bckg_intensities_fwhm(parameters, x_data, y_data, dy_data, bck_degree, central_angles,\
    wlength, uvw_pars=None):
    "This function fits background polynomials, intensities of individual Bragg-peaks\
    and optionally UVW parameters based on x_data, y_data, dy_data. If uvw_pars is not empty,\
    it performs the fitting task without fitting the Cagliotti parameters"
    simulated, _ = bckg_intensities_fwhm_pattern(parameters, x_data, y_data, bck_degree,\
        central_angles, wlength, uvw_pars)
    value = np.sum((np.divide((simulated-y_data), dy_data, np.zeros_like(dy_data),\
         where=dy_data != 0))**2)
    print value
    return value

def pattern_braggs(intensities, x_values, central_angles, uvw_pars):
    """This function creates a Bragg-pattern for central angles peak positions having uvw
    for the x_values"""
    pattern = np.zeros_like(x_values)
    for i, item in enumerate(central_angles):
        pattern += abs(intensities[i])*calpattoneref_uvw(x_values, item, uvw_pars)
    return pattern

def create_ratio(parameters, measured_datasets, central_angles, uvw_pars):
    """This function returns with the ratio of two simulated datasets determined by the parameters.
    Parameters contain the intensities of each Bragg-peaks and the background polynomials.
    Other input datasets are tuple of x1_data, x2_data, y_data1/y_data2, dy_data for (y1/y2),
    tuples of central angles, finally Cagliotti UVW."""
    start = 0
    stop = len(central_angles[0])
    simulated1 = 1+pattern_braggs(parameters[start:stop], measured_datasets[0], central_angles[0],\
        uvw_pars)
    start = stop
    stop = start + len(central_angles[1])
    simulated2 = 1+pattern_braggs(parameters[start:stop], measured_datasets[1], central_angles[1],\
        uvw_pars)
    back = polyn.Polynomial(parameters[stop:])
    return simulated1/simulated2 - back(measured_datasets[0])

def fit_vanadium_double_intensities(parameters, measured_datasets, central_angles, uvw_pars):
    """This function returns with the cost function between calculated and measured data.
    Parameters contain the intensities of each Bragg-peaks and the background polynomials.
    Other input datasets are tuple of x1_data, x2_data, y_data1/y_data2, dy_data for (y1/y2),
    tuples of central angles, finally Cagliotti UVW."""
    #Creating pattern based on the parameters
    simulated = create_ratio(parameters, measured_datasets, central_angles, uvw_pars)
    return np.sum((np.divide((measured_datasets[2]-simulated), measured_datasets[3],\
        np.zeros_like(simulated), where=measured_datasets[3] != 0))**2)

def main(argv):
    "Simulates the total diffraction pattern of vanadium"
    #np.set_printoptions(threshold='nan')
    if len(argv) < 12 or len(argv) > 15:
        print "This program creates efficiency file from vanadium dataset"
        print "\tUsage: "+argv[0]+" vanadium_ndt background_ndt"
        print "\t\tCagliotti_U V W wavelength[A] lattice_constant[A]"
        print "\t\tmean_square_displacement[A**2] coherent_xs[barn] incoherent_xs[barn]"
        print "\t\ttask_indicator[1-simulation/2-shifted_fit/3-plexi_fit/4-fit_by_itsetself]"
        print "[2-shifted_fit: vanadium_2ndt background_2ndt bckg_degree; 3-plexi_fit: plexi;"
        print " 4-fit_by_itself: bckg_poly_degree>=0 SavGol_poly_order<0]"
        print " [4-fit_by_itself: Savgol_interval]"
        exit()
    else:
        data_vanadium = np.loadtxt(argv[1]+".dat", unpack=True)
        data_back = np.loadtxt(argv[2]+".dat", unpack=True)
        data_net = data_vanadium[1] - data_back[1]
        data_dnet = np.sqrt(data_vanadium[2]**2+data_back[2]**2)
        number_density = 2./float(argv[7])**3
        brcorr = 1.0/number_density/float(argv[10])
        if int(argv[11]) == 1:
            #create Bragg pattern
            full_pattern = create_bragg_pattern(data_vanadium[3],\
                np.array([float(argv[3]), float(argv[4]), float(argv[5])]), float(argv[6]),\
                    float(argv[7]), float(argv[8]))
            #scaling it according to incoherent and coherent scattering
            full_pattern = (full_pattern*float(argv[9])+float(argv[10]))/4/np.pi
            np.savetxt(argv[1]+'_bragg.dat', np.c_[data_vanadium[4], full_pattern, data_net,\
                data_dnet, data_vanadium[3]], fmt='%f %f %e %e %f',\
                    header='Channel, Simulated_pattern, vanadium, dvanadium, degrees')
            return
        elif int(argv[11]) == 2:
            data_vanadium2 = np.loadtxt(argv[12]+".dat", unpack=True)
            data_back2 = np.loadtxt(argv[13]+".dat", unpack=True)
            data_net2 = data_vanadium2[1] - data_back2[1]
            data_dnet2 = np.sqrt(data_vanadium2[2]**2+data_back2[2]**2)
            degree = int(argv[14])
            #using only the parts, which is in the calibrated range
            lims = np.zeros(2, dtype=int)
            limits = np.where(data_vanadium[5] > 0.)
            lims[0] = limits[0][0]
            lims[1] = limits[0][-1] + 1
            data_rate = np.divide(data_net[lims[0]:lims[1]], data_net2[lims[0]:lims[1]],\
                    np.ones_like(data_net2[lims[0]:lims[1]]),\
                        where=data_net2[lims[0]:lims[1]] != 0)
            ddata_rate = (np.divide(data_dnet[lims[0]:lims[1]], data_net[lims[0]:lims[1]],\
                    np.zeros_like(data_net[lims[0]:lims[1]]),\
                        where=data_net[lims[0]:lims[1]] != 0.))**2
            ddata_rate += (np.divide(data_dnet2[lims[0]:lims[1]], data_net2[lims[0]:lims[1]],\
                    np.zeros_like(data_net2[lims[0]:lims[1]]),\
                        where=data_net2[lims[0]:lims[1]] != 0.))**2
            ddata_rate = np.sqrt(ddata_rate) * data_rate
            datasets = (data_vanadium[3][lims[0]:lims[1]],\
                data_vanadium2[3][lims[0]:lims[1]], data_rate, ddata_rate)
            peak_params1 = angleintensity_generator_bcc(float(argv[6]), float(argv[7]),\
                min(np.amax(data_vanadium[3][lims[0]:lims[1]]), 180.),\
                    max(0., np.amin(data_vanadium[3][lims[0]:lims[1]])))
            peak_params2 = angleintensity_generator_bcc(float(argv[6]), float(argv[7]),\
                min(np.amax(data_vanadium2[3][lims[0]:lims[1]]), 180.),\
                    max(0., np.amin(data_vanadium2[3][lims[0]:lims[1]])))
            parameters = []
            for element in peak_params1:
                parameters.append(1.0)
            for element in peak_params2:
                parameters.append(1.0)
            if degree >= 0:
                for i in range(degree+1):
                    parameters.append(0.0)
            result = minimize(fit_vanadium_double_intensities, parameters, args=(datasets,\
                (peak_params1, peak_params2), np.array([float(argv[3]), float(argv[4]),\
                    float(argv[5])])), method='Powell', options={'maxiter':2000000000})
            print result.message
            print "Final chi^2: ", result.fun
            parameters = result.x
            start = 0
            print "Peak intensities for pattern1: "
            for i in range(len(peak_params1)):
                print "\t"+str(i)+"th: "+str(abs(parameters[start+i]))
            start += len(peak_params1)
            print "Peak intensities for pattern2: "
            for i in range(len(peak_params2)):
                print "\t"+str(i)+"th: "+str(abs(parameters[start+i]))
            start += len(peak_params2)
            if degree >= 0:
                print "Background parameters: "
                for i in range(int(argv[14])+1):
                    print "\t"+str(i)+"th order: "+str(parameters[i])
            simulated_ratio = create_ratio(parameters, datasets, (peak_params1, peak_params2),\
                np.array([float(argv[3]), float(argv[4]), float(argv[5])]))
            simulated_pattern1 = 1+pattern_braggs(parameters[:len(peak_params1)], datasets[0],\
                peak_params1, np.array([float(argv[3]), float(argv[4]), float(argv[5])]))
            simulated_pattern2 = 1+pattern_braggs(parameters[len(peak_params1):\
                len(peak_params1)+len(peak_params2)], datasets[1], peak_params2,\
                    np.array([float(argv[3]), float(argv[4]), float(argv[5])]))
            simulated_pattern1 = np.pad(simulated_pattern1, (lims[0],\
                len(data_vanadium[0])-lims[1]), 'constant', constant_values=(1, 1))
            simulated_pattern2 = np.pad(simulated_pattern2, (lims[0],\
                len(data_vanadium[0])-lims[1]), 'constant', constant_values=(1, 1))
            data_rate = np.pad(data_rate, (lims[0], len(data_vanadium[0])-lims[1]), 'constant',\
                constant_values=(0, 0))
            simulated_ratio = np.pad(simulated_ratio, (lims[0], len(data_vanadium[0])-lims[1]),\
                'constant', constant_values=(0, 0))
            np.savetxt(argv[1]+'_efficiency.dat', np.c_[data_vanadium[4],\
                data_net/simulated_pattern1*brcorr, data_dnet*brcorr, data_rate, simulated_ratio,\
                    simulated_pattern1], fmt='%d %e %f %f %f %f',\
                        header='Channel_no, Bragg_filtered, err, measured_rate,'+\
                            ' fitted_rate, simulated_pattern')
            np.savetxt(argv[12]+'_efficiency.dat', np.c_[data_vanadium2[4],\
                data_net2/simulated_pattern2*brcorr, data_dnet2*brcorr, data_rate,\
                    simulated_ratio, simulated_pattern2], fmt='%d %e %f %f %f %f',\
                        header='Channel, Bragg_filtered, err, measured_rate,'+\
                            ' fitted_rate, simulated_pattern')
            return
        elif int(argv[11]) == 3:
            print "This option (plexi_fit) is not implemented yet!"
            return
        elif int(argv[11]) == 4:
            print "You have choosen fit to vanadium data by itself!"
            #Create initial dataset
            degree = int(argv[12])
            parameters = []
            multiplicator = np.sum(data_net)/len(data_net)
            if degree >= 0:
                for i in range(int(argv[12])+1):
                    parameters.append(0.0)
                parameters[0] = multiplicator
            peak_params = angleintensity_generator_bcc(float(argv[6]), float(argv[7]),\
                min(np.amax(data_vanadium[3]), 180.))
            peak_centers = np.zeros(len(peak_params))
            for i, item in enumerate(peak_params):
                peak_centers[i] = item
                parameters.append(peak_params.get(item)*multiplicator*\
                    float(argv[9])/float(argv[10]))
            wlength = 0
            if degree < 0:
                wlength = int(argv[13])
            #Optimization
            result = minimize(fit_bckg_intensities_fwhm, parameters, args=(data_vanadium[3],\
                data_net, data_dnet, int(argv[12]), peak_centers, wlength,\
                    np.array([float(argv[3]), float(argv[4]), float(argv[5])])), method='Powell',\
                                                            options={'maxiter':2000000000})
            print result.message
            print "Final chi^2: ", result.fun
            parameters = result.x
            start = 0
            if degree >= 0:
                print "Background parameters: "
                for i in range(int(argv[12])+1):
                    print "\t"+str(i)+"th order: "+str(parameters[i])
                    start += 1
            print "Peak intensities: "
            for i in range(len(peak_params)):
                print "\t"+str(i)+"th: "+str(abs(parameters[start+i]))
            calc_pattern, back_pattern = bckg_intensities_fwhm_pattern(parameters,\
                data_vanadium[3], data_net, int(argv[12]), peak_centers, wlength,\
                    np.array([float(argv[3]), float(argv[4]), float(argv[5])]))
            np.savetxt(argv[1]+'_bragg.dat', np.c_[data_vanadium[4],\
                data_net-calc_pattern+back_pattern, calc_pattern, back_pattern, data_net,\
                    data_dnet, data_vanadium[3]], fmt='%e %e %e %e %e %e %e',\
                        header='Channel, Bragg-peaks removed pattern, simulated_pattern,'+\
                            ' background pattern, vanadium, dvanadium, degrees')
            return

if __name__ == "__main__":
    main(sys.argv)
