#!/usr/bin/python
"This program merges the provided datasets"
import sys
import os.path
import numpy as np

class DatasetManipulation(object):
    "This class performs data manipilation"
    def __init__(self, with_errorbar=True):
        "Initialization: if errorbar is present, that value will be also read"
        self.x_data = np.empty(shape=(0, 0))
        self.y_data = np.empty(shape=(0, 0))
        self.dy_data = None
        if with_errorbar:
            self.dy_data = np.empty(shape=(0, 0))
        return

    def append(self, x_vals, y_vals, dy_vals=None):
        "Append the datasets into the structure."
        self.x_data = np.append(self.x_data, x_vals)
        self.y_data = np.append(self.y_data, y_vals)
        if dy_vals is not None and self.dy_data is not None:
            self.dy_data = np.append(self.dy_data, dy_vals)
        return

    def merge(self, resolution=-1.0, shift=0.5):
        "Merges the x datasets to bins of n*resolution..(n+1)*resolution if resolution>0.\
        The shift parameter indicates the x value related to the bin (0: lo, 0.5: central, 1: up).\
        If a bin do not contains a value, it will be skipped. If resolution is less than zero,\
        it merges only the matching x numbers."
        self.sort()
        if resolution <= 0.:
            i = 0
            while i < self.x_data.size:
                self.__merge_indices(np.argwhere(self.x_data == self.x_data[i]))
                i = i+1
        else:
            xdata = self.x_data/resolution
            xdata = xdata.astype(int)
            multi = False
            for i in range(np.amax(xdata)):
                if multi:
                    xdata = self.x_data/resolution
                    xdata = xdata.astype(int)
                    multi = False
                indices2merge = np.argwhere(xdata == i)
                if indices2merge.size != 0:
                    self.x_data[indices2merge[0]] = (i+shift)*resolution
                    if indices2merge.size > 1:
                        self.__merge_indices(indices2merge)
                        multi = True
        return

    def __merge_indices(self, idx_list):
        "Merges to datasets provided in idx_list into one point"
        if idx_list.size < 2:
            return
        if self.dy_data is not None:
            weights = 1./self.dy_data[idx_list]**2
            isum2 = np.sum(weights)
            self.y_data[idx_list[0]] = np.sum(self.y_data[idx_list]*weights)/isum2
            self.dy_data[idx_list[0]] = np.sqrt(1./isum2)
            self.dy_data = np.delete(self.dy_data, idx_list[1:])
        else:
            self.y_data[idx_list[0]] = np.sum(self.y_data[idx_list])/idx_list.size
        self.y_data = np.delete(self.y_data, idx_list[1:])
        self.x_data = np.delete(self.x_data, idx_list[1:])
        return

    def sort(self):
        "Sorts the datasets to increasing order"
        arr_idx = self.x_data.argsort()
        self.x_data = self.x_data[arr_idx]
        self.y_data = self.y_data[arr_idx]
        if self.dy_data is not None:
            self.dy_data = self.dy_data[arr_idx]
        return

    def write2file(self, fname):
        "Writes the datasets into fname file"
        data = [self.x_data] + [self.y_data]
        if self.dy_data is not None:
            data += [self.dy_data]
        np.savetxt(fname, np.column_stack(data))
        return

def main(argv):
    "Merges the provided datasets"
    if len(argv) < 8:
        print "This program merges experimental datasets and writes the unified to file"
        print "Usage: ", argv[0], " x_col_no y_col_no dy_col_no[-: not present]"+\
            "binwidth[-: exact match] bin_center_shift[0:lo, .5:mid, 1:hi]  outfile.name"+\
                " datasets.name"
    else:
        print "Data will read from columns:"
        if argv[3] == '-':
            selectcol = (int(argv[1])-1, int(argv[2])-1)
        else:
            selectcol = (int(argv[1])-1, int(argv[2])-1, int(argv[3])-1)
        print "\tX: "+str(selectcol[0]+1)
        print "\tY: "+str(selectcol[1]+1)
        if argv[3] == '-':
            print "\tdY: none"
        else:
            print "\tdY: "+str(selectcol[2]+1)
        print "After reading the datasets, merged file will be saved to \""+argv[6]+"\" file!"
        dataset = DatasetManipulation(argv[3] != '-')
        for _, fname in enumerate(argv[7:]):
            if not os.path.isfile(fname):
                print "ERROR: the \""+fname+"\" file does not exist!"
                return
            else:
                dyvals = None
                if argv[3] != '-':
                    xvals, yvals, dyvals = np.loadtxt(fname, usecols=selectcol, unpack=True)
                else:
                    xvals, yvals = np.loadtxt(fname, usecols=selectcol, unpack=True)
                dataset.append(xvals, yvals, dyvals)
                print "Datasets from file \""+fname+"\" has been reading in"
        if argv[4] == '-':
            print "Dataset will be sorted and merged when x-points exactly matched!"
            dataset.merge()
        else:
            print "The datasets are merged according to bins. Bin width: "+argv[4]+\
                "\n\tposition of 'x' value within the bin: resolution*"+argv[5]
            dataset.merge(float(argv[4]), float(argv[5]))
        dataset.write2file(argv[6])
        print "Merged dataset has been writing to \""+argv[6]+"\" file!"
    return

if __name__ == "__main__":
    main(sys.argv)
