#!/usr/bin/python
"The present program performs the first step of the calibration: from masked datasets determines\
the polynomial coeffs of the channel->tube polynomials, and the central channel"

################################################################
#Program has been created by Laszlo Temleitner on 2018 March
#License is granted for academic and non-commertial use.
#Version history:
# - positive intensity output is added (04/Feb/2020)
################################################################


from math import pi, sqrt
from os.path import basename, splitext, isfile
import sys
from scipy.optimize import minimize
#from scipy.signal import find_peaks_cwt
from combine_datasets import MTESTPSDData
import numpy as np

def fitt_mask(xval, space, parameters):
    "Calculates the masked intensity in channel space: parameters are [0] channel belonging to the\
    central gap, [1] intensities of each peaks, [2] channel to length polys, [3] sigma polys,\
    [4] background polynomials; if present: [5] calibration switch channel, [6] channel to length\
    polys upper, [7] sigma polys upper; xval are x coordinates and the gaps are spaced by space mm"
    value = np.zeros(xval.shape)
    times = 1
    coeffs = [parameters[2][::-1]]
    coeffs.append(np.zeros(1))
    ch2l = np.poly1d(np.concatenate(coeffs).ravel())
    center = float(parameters[0])
    sig = np.poly1d(parameters[3][::-1])
    bckg = np.poly1d(parameters[4][::-1])
    if len(parameters) > 6:
        times = 2
        coeffs2 = [parameters[6][::-1]]
        coeffs2.append(np.zeros(1))
        ch2l2 = np.poly1d(np.concatenate(coeffs2).ravel())
    for outer in range(times):
        low = int(round(xval[0]))-1
        high = int(round(xval[-1]))
        if times == 2:
            if outer == 0:
                high = int(round(parameters[5]))
            else:
                low = int(round(parameters[5]))
                ch2l = ch2l2
                if len(parameters) == 8:
                    sig = np.poly1d(parameters[7][::-1])
        for i, intensity in enumerate(parameters[1]):
            #Calculating the contribution of each Gaussian to the actual intensity
            l_sp = (i-len(parameters[1])/2)*space
            s_sp = sig(l_sp)
            value[low:high] += abs(intensity)*np.exp(-0.5*((ch2l(xval[low:high]-center)-l_sp)/\
                abs(s_sp))**2)/abs(s_sp)/sqrt(2*pi)
    value += bckg(xval-center)
    return value

def mask_min_expression(parameters, channel, yset, dyset, minch, maxch, space, no_peaks,\
                        ch2ldegree, sigdegree, backdegree):
    "Calculates the expression to be minimised: parameters are [0] channel belonging to the\
    central gap, [1] intensities of each peaks, [2] channel to length polys, [3] sigma polys,\
    [4] background polynomials; if present: [5] calibration switch channel, [6] channel to length\
    polys upper, [7] sigma polys upper\
    channel is x coordinates, ysets are y and dyset is dy, calculation will be performed between\
    minch and maxch channels (start form 1 is supposed) and the gaps are spaced by space mm,\
    additionally, no of peaks, the degree of channal-length, sigma and background polynomials\
    should be provided"
    params = [parameters[0]]
    params.append(np.array(parameters[1:no_peaks+1]))
    params.append(np.array(parameters[no_peaks+1:no_peaks+ch2ldegree+1]))
    params.append(np.array(parameters[no_peaks+ch2ldegree+1:no_peaks+ch2ldegree+sigdegree+2]))
    params.append(np.array(parameters[no_peaks+ch2ldegree+sigdegree+2:\
        no_peaks+ch2ldegree+sigdegree+backdegree+3]))
    if len(parameters) > no_peaks+ch2ldegree+sigdegree+backdegree+3:
        params.append(parameters[no_peaks+ch2ldegree+sigdegree+backdegree+3])
        params.append(np.array(parameters[no_peaks+ch2ldegree+sigdegree+backdegree+4:\
            no_peaks+2*ch2ldegree+sigdegree+backdegree+4]))
        if len(parameters) > no_peaks+2*ch2ldegree+sigdegree+backdegree+4:
            params.append(np.array(parameters[no_peaks+2*ch2ldegree+sigdegree+backdegree+4:\
                no_peaks+2*ch2ldegree+2*sigdegree+backdegree+5]))
    y_term = fitt_mask(channel, space, params)
    np.set_printoptions(threshold=1000)
    #print params
    #np.savetxt("ki.txt", np.column_stack([channel]+[y_term]+[yset]))
    sumterm = np.sum(((y_term[minch-1:maxch]-yset[minch-1:maxch])/dyset[minch-1:maxch])**2)
    if len(params) > 5:
        val = round(params[5])-params[0]
        coeffs = [params[2][::-1]]
        coeffs.append(np.zeros(1))
        ch2l = np.poly1d(np.concatenate(coeffs).ravel())
        coeffs2 = [params[6][::-1]]
        coeffs2.append(np.zeros(1))
        ch2l2 = np.poly1d(np.concatenate(coeffs2).ravel())
        sumterm += (ch2l2(val)-ch2l(val))**2
    return sumterm

def single_gauss_min(parameters, x_vals, y_vals):
    "Calculates the expression to be minimised: single Gaussian with constant background,\
    parameters are intensity, center, sigma, background"
    y_term = parameters[0]/sqrt(2*pi)/parameters[2]*\
        np.exp(-((x_vals-parameters[1])/parameters[2])**2/2.)+parameters[3]
    return np.sum((y_term-y_vals)**2)

class MTESTCalibrationChannelPosition(object):
    "This class describes the first part of the MTEST calibration as determining the electrical\
    channel - position relationship"
    def __init__(self, masked_vanadium_bckg, min_channels, max_channels, mm_between_holes):
        "Initializes the class by definying the filenames of the masked,\
        the corresponding vanadium and background measurements, then for each individual PSD\
        detectors: minimal; then maximal channel numbers where calibration is valid; finally\
        spacing between centre of gaps in the mask"
        self.segno = 0
        while isfile(masked_vanadium_bckg[0]+'_'+str(self.segno+1)+'_ndt.dat'):
            self.segno +=1
        for i in range(self.segno):
            #normalising the data and writing it to corrected file
            masked = np.loadtxt(splitext(basename(masked_vanadium_bckg[0]))[0]+"_"+str(i+1)+\
                "_ndt.dat", unpack=True)
            vanadium = np.loadtxt(splitext(basename(masked_vanadium_bckg[1]))[0]+"_"+str(i+1)+\
                "_ndt.dat", unpack=True)
            bckg = np.loadtxt(splitext(basename(masked_vanadium_bckg[2]))[0]+"_"+str(i+1)+\
                "_ndt.dat", unpack=True)
            corr = np.subtract(vanadium[1], bckg[1])
            corre = np.sqrt(np.power(vanadium[2], 2) + np.power(bckg[2], 2))
            resul = np.divide(masked[1], corr, np.zeros_like(masked[1]), where=corr != 0)
            resule = np.sqrt(np.power(np.divide(masked[2], masked[1], np.zeros_like(masked[2]),\
                where=masked[1] != 0), 2) + np.power(np.divide(corre, corr, np.zeros_like(corre),\
                where=corr != 0), 2))
            resule = np.multiply(resule, resul)
            dataset = [masked[4]] + [resul] + [resule]
            np.savetxt("masked_corr_"+str(i+1)+".dat", np.column_stack(dataset), header=\
                ' channel, Corr, dCorr')
        self.spacing = mm_between_holes
        self.mins = min_channels
        self.maxs = max_channels
        return

    def get_calibration_parameters(self, calibration_codes, ch2mm_coeffs_dimension,\
        sigma_dimension, bckg_dimension, list_of_peaks, central_peak_idx, outfile):
        "Determines the calibration parameters"
        fil = open(outfile, "w")
        print "Calibration file \"", outfile, "\" is creating"
        fil.write("%d !#tubes\n" % (len(self.mins)))
        fil.write("ni ni_bckg van van_bckg !for each positions ndt files for: nickel,")
        fil.write(" background of nickel, vanadium, background of vanadium\n")
        fil.write("1 1 ?? !#distance_sample-tube #height #detector_banks\n")
        fil.write("825.4 !common_distances[mm]\n35. !common heights[mm]\n")
        fil.write("?? !angular shifts of each detector banks\n")
        for i in range(self.segno):
            if calibration_codes[i] != 0:
                data = np.loadtxt("masked_corr_"+str(i+1)+".dat", unpack=True)
                #Determining the initial parameters: the channel-distance linear parameter
                no_peaks, center, ch2poly1, intens, sig0, bckg0 =\
                    self.__determine_initial_parameters(data[1], i, list_of_peaks,\
                        central_peak_idx)
                print "Detector no ", i+1
                print "\tInitial optimalization found ", no_peaks,\
                    " number of peaks with central peak at channel ", center
                print "\tThe channel2detector linear coefficient is ", ch2poly1,\
                    " mm, intensity: ", intens, ", sigma: ", sig0
                print "\tUp to "+str(ch2mm_coeffs_dimension[i])+\
                    "th order channel->detector coeffs will be fitted!"
                print "\tUp to "+str(sigma_dimension[i])+\
                    "th order sigma polynomials will be fitted!"
                print "\tUp to "+str(bckg_dimension[i])+\
                    "th order background polynomials will be fitted!"
                params = [np.array([center]), np.full(no_peaks, intens),\
                    np.zeros(ch2mm_coeffs_dimension[i]),\
                    np.zeros(sigma_dimension[i]+1),\
                        np.zeros(bckg_dimension[i]+1)]
                params[2][0] = ch2poly1
                params[3][0] = sig0
                params[4][0] = bckg0
                if calibration_codes[i] == 2:
                    params.extend([params[0], params[2], params[3]])
                    print "\tThe valid range is splitted, calibration barrier will be fitted"
                elif calibration_codes[i] == 3:
                    params.extend([params[0], params[2]])
                    print "\tThe valid range is splitted, calibration barrier will be fitted"+\
                        ", but with common sigma"
                result = minimize(mask_min_expression, np.concatenate(params).tolist(),\
                    args=(data[0], data[1], data[2], self.mins[i], self.maxs[i], self.spacing[i],\
                        no_peaks, ch2mm_coeffs_dimension[i], sigma_dimension[i],\
                            bckg_dimension[i]), method='Powell')
                if result.success:
                    high = 0
                    par = []
                    for _, k in enumerate(params):
                        low = high
                        high += len(k)
                        par.append(np.array(result.x[low:high]))
                    print 'The calibration procedure for detector no '+str(i+1)+\
                        ' has been finished properly. Chi^2: '+str(result.fun)
                    print '\tCenter channel no: '+str(par[0])
                    for j, k in enumerate(par[1]):
                        print '\tIntensity of gap '+str(j-len(par[1])/2)+' is '+str(abs(k))
                    print '\tThe coefficients for channel->length conversion polynomial:'
                    for j, k in enumerate(par[2]):
                        print '\t\t'+str(j+1)+'th order: '+str(k)
                    print '\tThe coefficients for sigma polynomial:'
                    for j, k in enumerate(par[3]):
                        print '\t\t'+str(j)+'th order: '+str(k)
                    print '\tThe coefficients for background polynomial:'
                    for j, k in enumerate(par[4]):
                        print '\t\t'+str(j)+'th order: '+str(k)
                    if len(par) > 5:
                        print '\tCalibration barrier at channel '+str(par[5])
                        print '\tThe coefficients for upper channel->length conversion polynomial:'
                        for j, k in enumerate(par[6]):
                            print '\t\t'+str(j+1)+'th order: '+str(k)
                        if len(par) > 7:
                            print '\tThe coefficients for upper sigma polynomial:'
                            for j, k in enumerate(par[7]):
                                print '\t\t'+str(j)+'th order: '+str(k)
                    np.savetxt("masked_corr_fitted_"+str(i+1)+".dat",\
                        np.column_stack([data[0]]+[fitt_mask(data[0], self.spacing[i], par)]+\
                            [data[1]]+[data[2]]), header='channel fit norm dnorm')
                    if len(par) > 5:
                        fil.write("2 %d %d %d " % (self.mins[i], int(par[5]), self.maxs[i]))
                    else:
                        fil.write("1 %d %d " % (self.mins[i], self.maxs[i]))
                    fil.write("%d %g " % (ch2mm_coeffs_dimension[i], par[0]))
                    for _, k in enumerate(par[2]):
                        fil.write("%g " % (k))
                    if len(par) > 5:
                        for _, k in enumerate(par[6]):
                            fil.write("%g " % (k))
                    fil.write("\n")
                else:
                    print 'The calibration procedure for detector no '+str(i+1)+' has been failed'
                    fil.write("Please, fill manually\n")
            else:
                fil.write("Please, fill manually\n")
        fil.close()
        return

    def __determine_initial_parameters(self, y_data, no_det, peak_list, central_peak_indx):
        "This function provides the initial parameters for the fitting procedure:\
        inputs arbckg_1_4e the y_data (x indices are assumed to start from 1), minimal, then the detector\
        number; --- It returns with: number of peaks in range, the central peak position in\
        channel, the channel2position first order coefficient in [mm] units assuming channel\
        number is provided in reduced units, the intensity, the 0th order sigma and bckg parameter"
        #peaks = find_peaks_cwt(y_data[self.mins[no_det]-1:self.maxs[no_det]], np.arange(5, 10))
        #peakpos = np.add(peaks, self.mins[no_det]-1)
        #peak_distance = float(peakpos[-2]-peakpos[1])/(len(peakpos)-3)
        #no_available_peaks = int((self.maxs[no_det]-self.mins[no_det]+1)/peak_distance)
        #if no_available_peaks%2 == 0:
        #    no_available_peaks -= 1
        #centre_candidates = np.where(np.logical_and(peakpos >= self.mins[no_det]+\
        #    (no_available_peaks-1)/2*peak_distance, peakpos <= self.maxs[no_det]-\
        #        (no_available_peaks-1)/2*peak_distance))
        #diffc = abs(peakpos[1]+((no_available_peaks-1)/2-1)*peak_distance -\
        #    peakpos[centre_candidates[0]])
        #center = peakpos[centre_candidates[0][np.argmin(diffc)]]
        #init_ch2pos1 = self.spacing[no_det]*(len(peakpos)-3)/float(peakpos[-2]-peakpos[1])
        #from now: new codes
        no_available_peaks = len(peak_list[no_det])
        center = peak_list[no_det][central_peak_indx[no_det]]
        peak_distance = peak_list[no_det][central_peak_indx[no_det]+1]-center
        init_ch2pos1 = (no_available_peaks-1)*self.spacing[no_det]/\
            (peak_list[no_det][-1]-peak_list[no_det][0])
        intensity = y_data[int(center)]*5
        sigma0 = 1.0/2.354820045*0.2*self.spacing[no_det]
        lims = [int(center-peak_distance/2.), int(center+peak_distance/2.)]
        result = minimize(single_gauss_min, [intensity, center, sigma0, 0.],\
                    args=(np.arange(lims[0], lims[1]), y_data[lims[0]:lims[1]]),\
                        method='Nelder-Mead')
        intensity = result.x[0]
        sigma0 = result.x[2]*init_ch2pos1
        bckg = result.x[3]
        return no_available_peaks, center, init_ch2pos1, intensity, sigma0, bckg

def main(argv):
    "Does the fit"
    if len(argv) != 3:
        print "This program performs the first part of calibration: the channel->"
        print "tube position calibration"
        print "\tUsage: "+argv[0]+" control_file.name output_calib.filename"
        print "The control file should contain the following items:"
        print "\t#tubes(incl. non-fitted tubes)"
        print "\tfilenames: masked[*_ndt.dat] vanadium[*_ndt.dat] vanadium_bckg[*_ndt.dat]"
        print "\tfor each tubes: calibration_code[omit(0)/fit single range(1)/double range(2)",\
            "/with common sigma(3)] min_channel max_channel distance_between_gaps[mm]",\
                "ch->poly_degree sigma_degree background_poly_degree"
        exit()
    with open(argv[1], 'r') as fil:
        print "Reading from the \"", argv[1], "\" control file:"
        ##tubes
        line = fil.next()
        splt = line.split(' ')
        no_tubes = int(splt[0])
        print "Number of tubes: ", no_tubes
        filenames = []
        line = fil.next()
        splt = line.split(' ')
        for i in range(3):
            filenames.append(splt[i])
        print "Filenames (masked, vanadium, vanadium_bckg): ", filenames
        calibcodes = []
        minimals = []
        maximals = []
        hole_distances = []
        degrees = []
        sdegrees = []
        bckg = []
        deadtimes = []
        peaks = []
        central_peak = np.zeros(no_tubes, dtype=np.int_)
        for i in range(no_tubes):
            print "Parameters for tube no. "+str(i+1)
            line = fil.next()
            splt = line.split(' ')
            calibcodes.append(int(splt[0]))
            if calibcodes[-1] == 0:
                print "\tTube will be omitted from fit"
            elif calibcodes[-1] == 1:
                print "\tParticipate in fit with one range"
            elif calibcodes[-1] == 2:
                print "\tParticipate in fit with two ranges, sigmas are also fitted separately"
            elif calibcodes[-1] == 3:
                print "\tParticipate in fit with two ranges, sigmas are common"
            else:
                print "\tERROR: invalid calibration code!"
                exit()
            minimals.append(int(splt[1]))
            maximals.append(int(splt[2]))
            print "\tValid range: ", minimals[-1], "...", maximals[-1], " channels"
            hole_distances.append(float(splt[3]))
            print "\tDistance between gaps: ", hole_distances[-1], "mm"
            degrees.append(int(splt[4]))
            sdegrees.append(int(splt[5]))
            bckg.append(int(splt[6]))
            print "\tPolynomial degree for:\n\t\tchannel->tube transformation: ", degrees[-1]
            print "\t\tsigma(tube): ", sdegrees[-1], "\t\tbackground: ", bckg[-1]
            no_peaks = int(splt[7])
            central_peak[i] = int(splt[8])-1
            print "\tNumber of peaks: ", no_peaks
            peak_items = np.zeros(no_peaks)
            for j in range(no_peaks):
                peak_items[j] = float(splt[9+j])
            peaks.append(peak_items)
            print "\tPeak positions: ", peaks[-1]
            print "\tCentral peak at ", peaks[-1][central_peak[i]]
        print "Initializing.."
        raw = MTESTCalibrationChannelPosition(filenames, minimals, maximals, hole_distances)
        print "Start to fit.."
        raw.get_calibration_parameters(calibcodes, degrees, sdegrees, bckg,\
            peaks, central_peak, argv[2])
    return

if __name__ == "__main__":
    main(sys.argv)
