#!/usr/bin/python
"The present program performs the last step of the calibration: determines wavelength, the\
Cagliotti UVW parameters, angular shift of each channels,  and some common distances"
from math import pi, cos, sin, radians, sqrt, log
from os.path import isfile
import sys
from scipy.optimize import minimize
import numpy as np
import numpy.polynomial.polynomial as polyn

def calc_degree_from_channel(channels, ch2pos, central_shift, distance, height, central_angle,\
    reference_channel, limits):
    "Calculates the degree for each channel positions. Inputs are channel positions, list of\
    channel->pos polynomials, perpendicular-central_point distance on tube, tube-sample distance,\
    height above the scattering plane, angle belonging to the central point, reference channel,\
    finally limits (min, limiting chns for each ch2pos polys, max_chn)"
    rel_h2 = (height/distance)**2
    rel_shift = central_shift/distance
    ccentral = cos(radians(central_angle))
    scentral = sin(radians(central_angle))
    rel_tube = np.zeros(limits[-1]-limits[0]+1)
    rel_tube[0:limits[1]-limits[0]+1] = ch2pos[0](channels[limits[0]-1:limits[1]]\
        -reference_channel)/distance
    for i in range(2, len(limits)):
        rel_tube[limits[i-1]-limits[0]+1:limits[i]-limits[0]+1] =\
            ch2pos[i-1](channels[limits[i-1]:limits[i]]-reference_channel)/distance
    nominator = np.sqrt(1.0+rel_h2+(rel_shift+rel_tube)**2)
    return np.degrees(np.arccos((ccentral-scentral*(rel_tube+rel_shift))/nominator))

def calc_model_pattern(wavelength, inverse_d2, intensities, uvw_poly, back_poly, degs):
    "Calculates the model pattern at degrees -- it returns with the corresponding intensity.\
    Inputs are wavelengths, the 1/d^2 array, intensities, fwhms belonging for each peaks or\
    Cagliotty UVW polynomial, the background polynomial and the array of degrees"
    peak_positions = np.degrees(np.arccos(1.0-0.5*inverse_d2*wavelength**2))
    #if (uvw_poly) is np.lib.polynomial.poly1d:
    if isinstance(uvw_poly, np.polynomial.polynomial.Polynomial):
        fwhms = create_fwhm_list(uvw_poly, peak_positions)
    else:
        fwhms = uvw_poly
    pos_intensity = np.zeros_like(degs)
    for i, ppos in enumerate(peak_positions):
        pos_intensity += np.abs(intensities[i]/fwhms[i])*2.*sqrt(log(2.)/pi)*\
            np.exp(-log(2.)*(2*(degs-ppos)/fwhms[i])**2)
    pos_intensity += back_poly(degs)
    return pos_intensity

def create_fwhm_list(uvw_poly, peak_positions):
    "Creates fwhm list belonging to peak_positions provided in degrees. U, V, W are the Cagliotti\
    parameters"
    tan_val = np.tan(np.radians(peak_positions)/2.)
    return np.sqrt(uvw_poly(tan_val))

def inverse_d2_between_planes(l_params, list_of_reflections):
    "Creates list of decreasing inverse square distances of lattice planes provided by list\
    of reflections and lattice_parameters"
    easyvol = l_params[0]*l_params[1]*l_params[2]
    volume = easyvol * sqrt(1.0-cos(radians(l_params[3]))**2-cos(radians(l_params[4]))**2-\
        cos(radians(l_params[5]))**2+2*cos(radians(l_params[3]))*cos(radians(l_params[4]))*\
            cos(radians(l_params[5])))
    mul = np.array([(easyvol/l_params[0]*sin(radians(l_params[3])))**2,\
        (easyvol/l_params[1]*sin(radians(l_params[4])))**2,\
            (easyvol/l_params[2]*sin(radians(l_params[5])))**2,\
                easyvol*l_params[2]*(cos(radians(l_params[3]))*cos(radians(l_params[4]))-\
                    cos(radians(l_params[5]))),\
                        easyvol*l_params[0]*(cos(radians(l_params[4]))*\
                            cos(radians(l_params[5]))-cos(radians(l_params[3]))),\
                                easyvol*l_params[1]*(cos(radians(l_params[5]))*\
                                    cos(radians(l_params[3]))-cos(radians(l_params[4])))])
    mul /= volume**2
    refl_list = np.zeros(len(list_of_reflections))
    for i, j in enumerate(list_of_reflections):
        refl = np.array([j[0]**2, j[1]**2, j[2]**2, 2*j[0]*j[1], 2*j[1]*j[2], 2*j[2]*j[0]])
        refl_list[i] = np.dot(refl, mul)
    return np.unique(refl_list.round(decimals=7))

def calc_segment_pattern_intensity(wavelength, inverse_d2, intensities, uvw_poly, back_poly,\
    channels, ch2pos, central_shift, distance, height, central_angle, reference_channel, limits):
    "Calculates the representation of the model pattern at one position of the tube. Inputs are\
    the wavelength, the 1/d^2 array, intensities belonging for each peaks, fwhm for each\
    peaks/Cagliotti UVW polynomial, background polynomial of the pattern, channels, the list of\
    pos->channel transformation polynomial for each range, perpendicular-central_point distance on\
    tube, tube-sample distance, height above the scattering plane, angle belonging to the central\
    point, reference channel, finally limits (min, limiting chns for each ch2pos polys, max_chn)"
    degs = calc_degree_from_channel(channels, ch2pos, central_shift,\
        distance, height, central_angle, reference_channel, limits)
    return calc_model_pattern(wavelength, inverse_d2, intensities, uvw_poly, back_poly, degs)

def system_min_expression(parameters, bckg_degree, inverse_d2, no_distances,\
    no_heights, no_ang_shifts, distance_idx, distance_shifts, height_idx, height_mult,\
        ang_shft_idx, ch2pos, ref_chn, limits, encoder_angles, x_data, y_data, dy_data,\
            each_fwhm_fit, wvu_fixed=None, wavelength=0.0, heights=None, distances=None,\
                angle_shifts=None, tube_center_shift=None):
    "Calculates the expression to be minimized for the calibration of the system. The parameters\
    contains the wavelength[0], {intensity, fwhm}_reflections, background_polynomials,\
    common distances, heights and shifts, {tube_center_shift}_each_tubes. Then\
    poly_bckg_degree, 1/d^2 values, #distances, #heights, #angle shifts, for\
    each tubes: distance indices, distance shifts, height indices, height multiplicator,\
    angle_shift_index, list of channel->position polynomials (it might be more than one set for a\
    tube), reference channel, lists of np array of limits (min, limiting chns for each ch2pos\
    polys, max_chn) of each tubes, positions: encoder angles; finally x, y, dy datasets should be\
    provided in {{}_position}_tube system should be provided"
    intensity = np.zeros_like(inverse_d2)
    fwhm_wvu = np.zeros(3)
    if each_fwhm_fit:
        fwhm_wvu = np.zeros_like(inverse_d2)
    no_fwhm_vary = len(fwhm_wvu)
    if wvu_fixed != None:
        no_fwhm_vary -= len(wvu_fixed)
        for i in range(no_fwhm_vary, 3):
            fwhm_wvu[i] = wvu_fixed[i-no_fwhm_vary]
    bckg_p = np.zeros(bckg_degree+1)
    if wavelength == 0.0 and distances is None and angle_shifts is None and\
        tube_center_shift is None:
        wavelength = parameters[0]
        distances = np.zeros(no_distances)
        for i, _ in enumerate(distances):
            distances[i] = parameters[2+bckg_degree+len(inverse_d2)+no_fwhm_vary+i]
        actual = 2+bckg_degree+len(inverse_d2)+no_fwhm_vary+no_distances
        if heights is None:
            heights = np.zeros(no_heights)
            for i, _ in enumerate(heights):
                heights[i] = parameters[actual+i]
            actual += no_heights
        if angle_shifts is None:
            angle_shifts = np.zeros(no_ang_shifts)
            for i, _ in enumerate(angle_shifts):
                angle_shifts[i] = parameters[actual+i]
            actual += no_ang_shifts
        tube_center_shift = np.zeros(len(distance_idx))
        for i, _ in enumerate(limits):
            tube_center_shift[i] = \
                parameters[actual+i]
        for i, _ in enumerate(intensity):
            intensity[i] = parameters[1+i]
#            print "\tpeak=", i+1, ", intensity=", intensity[i]
        for i in range(no_fwhm_vary):
            fwhm_wvu[i] = parameters[1+len(inverse_d2)+i]
        for i, _ in enumerate(bckg_p):
            bckg_p[i] = parameters[1+len(inverse_d2)+no_fwhm_vary+i]
    else:
        for i, _ in enumerate(intensity):
            intensity[i] = parameters[i]
#            print "\tpeak=", i+1, ", intensity=", intensity[i]
        for i in range(no_fwhm_vary):
            fwhm_wvu[i] = parameters[len(inverse_d2)+i]
        for i, _ in enumerate(bckg_p):
            bckg_p[i] = parameters[len(inverse_d2)+no_fwhm_vary+i]
    #bckg = np.poly1d(bckg_p[::-1])
    bckg = polyn.Polynomial(bckg_p)
    if each_fwhm_fit:
        uvw_poly = fwhm_wvu
    else:
        #uvw_poly = np.poly1d(fwhm_wvu[::-1])
        uvw_poly = polyn.Polynomial(fwhm_wvu)
#    print "lambda=", wavelength
#    print "Distances: ", distances
#    print "Heights: ", heights
#    for i, _ in enumerate(central_angle_shift):
#        print "\ttube=", i+1, ", central_angle_shift=", central_angle_shift[i],\
#                ", tube_center_shift=", tube_center_shift[i]
#    print "\tBackground:"
#    print bckg
#    print "\tFWHM parameters:"
#    print uvw_poly
    cost = 0.0
    for detector, detlim in enumerate(limits):
        for position, _ in enumerate(encoder_angles):
            pidx = detector*len(encoder_angles)+position
            y_calc = calc_segment_pattern_intensity(wavelength, inverse_d2, intensity, uvw_poly,\
                bckg, x_data[pidx],\
                    ch2pos[detector], tube_center_shift[detector],\
                        distances[distance_idx[detector]]+distance_shifts[detector],\
                            heights[height_idx[detector]]*height_mult[detector],\
                                angle_shifts[ang_shft_idx[detector]]+encoder_angles[position],\
                                    ref_chn[detector], detlim)
            cost += np.sum((np.divide(y_calc-y_data[pidx][(detlim[0]-1):detlim[-1]],\
                dy_data[pidx][(detlim[0]-1):detlim[-1]],\
                    np.zeros_like(dy_data[pidx][(detlim[0]-1):detlim[-1]]),\
                        where=dy_data[pidx][(detlim[0]-1):detlim[-1]] != 0))**2)
    #print cost
    return cost

class MTESTCalibrationSystem(object):
    "This class provides procedures for the final calibration of the MTEST detector system"
    def __init__(self, nickel_nibckg_vanadium_vanbckg, limit_channels, ch2pos_polys,\
        center):
        #TODO: to check
        #TODO: remove Bragg-peaks from the vanadium pattern
        "This method prepares the dataset to get corrected for the final calibration:\
        contains the filename of [0] nickel, [1] the corresponding backgrounds, [2] vanadium and\
        the corresponding [3] background. Lists of np array of limits (min, limiting chns for each\
        ch2pos polys, max_chn) of each tubes, the list of channel ->position coding polynomials,\
        the center position for each detectors"
        #Determining number of segments and detector positions
        index = 0
        tube_name = ""
        tube_no_looking = True
        tube_centre = []
        while isfile(nickel_nibckg_vanadium_vanbckg[0]+'_'+str(index+1)+'_ndt.dat'):
            with open(nickel_nibckg_vanadium_vanbckg[0]+'_'+str(index+1)+'_ndt.dat') as fil:
                line = fil.readline()
                splitted = line.split(' ')
                if index == 0:
                    tube_name = splitted[-1]
                    tube_centre.append(splitted[-3])
                elif splitted[-1] == tube_name:
                    if tube_no_looking:
                        tube_no_looking = False
                        self.no_detector_tubes = index
                    found = False
                    for _, i in enumerate(tube_centre):
                        if i == splitted[-3]:
                            found = True
                    if not found:
                        tube_centre.append(splitted[-3])
            index += 1
        #tube_centre = ','.join(tube_centre)
        self.ref_angles = np.fromstring(','.join(tube_centre), dtype=np.float, sep=',')
        for i in range(self.no_detector_tubes):
            for j, _ in enumerate(self.ref_angles):
                #normalising the data and writing it to corrected file
                nickel = np.loadtxt(nickel_nibckg_vanadium_vanbckg[0]+\
                    "_"+str(j*self.no_detector_tubes+i+1)+"_ndt.dat", unpack=True)
                nibckg = np.loadtxt(nickel_nibckg_vanadium_vanbckg[1]+\
                    "_"+str(j*self.no_detector_tubes+i+1)+"_ndt.dat", unpack=True)
                vanadi = np.loadtxt(nickel_nibckg_vanadium_vanbckg[2]+\
                    "_"+str(j*self.no_detector_tubes+i+1)+"_ndt.dat", unpack=True)
                vabckg = np.loadtxt(nickel_nibckg_vanadium_vanbckg[3]+\
                    "_"+str(j*self.no_detector_tubes+i+1)+"_ndt.dat", unpack=True)
                corr = np.subtract(vanadi[1], vabckg[1])
                corre = np.sqrt(np.power(vanadi[2], 2) + np.power(vabckg[2], 2))
                corrd = np.subtract(nickel[1], nibckg[1])
                corrde = np.sqrt(np.power(nickel[2], 2) + np.power(nibckg[2], 2))
                resul = np.divide(corrd, corr, np.zeros_like(corrd), where=corr != 0)
                resule = np.sqrt(np.power(np.divide(corrde, corrd, np.zeros_like(corrde),\
                    where=corrd != 0), 2) + np.power(np.divide(corre, corr, np.zeros_like(corre),\
                        where=corr != 0), 2))
                resule = np.multiply(resule, resul)
                dataset = [nickel[4]] + [resul] + [resule]
                np.savetxt("nickel_corr_"+str(j+1)+"_"+str(i+1)+".dat", np.column_stack(dataset),\
                    header=' channel, Corr, dCorr')
        self.ch2pos = ch2pos_polys
        self.limits = limit_channels
        self.center = center
        return

    def do_calibration(self, wavelength, lattice_parameters, list_of_reflections, poly_deg,\
        tube_sample_distances_mm, tube_sample_distance_indices, tube_sample_shift,\
            height_above_plane_mm, tube_height_indices, tube_height_multiplicator,\
                ttheta_shifts, ttheta_shift_indices, detector_indx2fit,\
                    out_calibration):
        "Performs calibration of the whole system. Parameters are wavelength in AA; lattice\
        constants and angles; list of reflections; polynomial degree of background, common list\
        of sample-tube distances in mm, indices of the previous for each tubes,  then shift from\
        ideal values; common tube heights above the scattering plane in mm, height indices for\
        each tubes, then the applied multiplicator; 2theta shift of shortest distance angle in\
        comparison with the motor coded angle (common for each banks); 2theta shift indices;\
        indicating that the detector takes apart in the calibration; output calibration filename"
        #create list of distances between lattice planes
        invd2_list = inverse_d2_between_planes(lattice_parameters, list_of_reflections)
        #filter out non-relevant reflections
        invlist = invd2_list*wavelength**2
        initial_list = invlist[invlist < 4.]
        initial_list = np.degrees(np.arccos(1.0-0.5*initial_list))
        #In the first stage, rough initialisation will done: intensity, fwhm, bckg_poly are refined
        parameters_easy = []
        for i, _ in enumerate(initial_list):
            parameters_easy.append(10.)
        for i, _ in enumerate(initial_list):
            parameters_easy.append(.5)
        for i in range(poly_deg+1):
            parameters_easy.append(0.0)
        ch2pos_restricted = []
        limits_restricted = []
        x_data = []
        y_data = []
        dy_data = []
        for i, val in enumerate(detector_indx2fit):
            if val:
                ch2pos_restricted.append(self.ch2pos[i])
                limits_restricted.append(self.limits[i])
                for position, _ in enumerate(self.ref_angles):
                    data = np.loadtxt("nickel_corr_"+str(position+1)+"_"+str(i+1)+".dat",\
                        unpack=True)
                    x_data.append(data[0])
                    y_data.append(data[1])
                    dy_data.append(data[2])
        result = minimize(system_min_expression, parameters_easy, args=(poly_deg,\
            invd2_list[:len(initial_list)], len(tube_sample_distances_mm),\
                len(height_above_plane_mm), len(ttheta_shifts),\
                    tube_sample_distance_indices[detector_indx2fit],\
                        tube_sample_shift[detector_indx2fit],\
                            tube_height_indices[detector_indx2fit],\
                                tube_height_multiplicator[detector_indx2fit],\
                                    ttheta_shift_indices[detector_indx2fit], ch2pos_restricted,\
                                        self.center[detector_indx2fit], limits_restricted,\
                                            self.ref_angles, x_data, y_data, dy_data, True,\
                                            None, wavelength, height_above_plane_mm,\
                                                tube_sample_distances_mm,\
                                                    ttheta_shifts,\
                                                        np.zeros(np.sum(detector_indx2fit))),\
                                                            method='Powell')
        print "Initial stage has been finished (peak params init). Chi^2: ", result.fun
        #In the second stage, wavelength, intensity_each, fwhm_each, bckg_poly,
        # tube_sample_distances, 2theta_bank_shifts, tube_perpendicular_shifts are refined
        parameters = []
        parameters.append(wavelength)
        for j, i in enumerate(result.x):
            result.x[j] = abs(i)
        parameters.extend(result.x)
        #parameters.extend(result.x[:(len(result.x)-poly_deg-1)])
        #parameters.append(1.0)
        #parameters.append(0.0)
        #parameters.append(0.0)
        #parameters.extend(result.x[(len(result.x)-poly_deg-1):])
        for _, i in enumerate(tube_sample_distances_mm):
            parameters.append(i)
        #for _, i in enumerate(height_above_plane_mm):
        #    parameters.append(i)
        for _, i in enumerate(ttheta_shifts):
            parameters.append(i)
        for i, val in enumerate(detector_indx2fit):
            if val:
                #parameters.append(ttheta_shift_for_each[i])
                parameters.append(0.0)
        result = minimize(system_min_expression, parameters, args=(poly_deg,\
            invd2_list[:len(initial_list)], len(tube_sample_distances_mm),\
                 len(height_above_plane_mm), len(ttheta_shifts),\
                    tube_sample_distance_indices[detector_indx2fit],\
                        tube_sample_shift[detector_indx2fit],\
                            tube_height_indices[detector_indx2fit],\
                                tube_height_multiplicator[detector_indx2fit],\
                                    ttheta_shift_indices[detector_indx2fit],\
                                        ch2pos_restricted, self.center[detector_indx2fit],\
                                            limits_restricted,\
                                                self.ref_angles, x_data, y_data, dy_data, True,\
                                                    None, 0.0, height_above_plane_mm),\
                                                        method='Nelder-Mead',\
                                                            options={'maxiter':2000000000})
        print "1/4 has been finished (individual FWHMs have been fitted). Chi^2: ", result.fun
        fwhm_par = polyn.polyfit(np.tan(\
            np.arccos(1.0-0.5*invd2_list[:len(initial_list)]*result.x[0]**2)/2.),\
             (result.x[len(initial_list)+1:2*len(initial_list)+1])**2, 2,\
                w=np.divide(result.x[1:len(initial_list)+1],\
                 result.x[len(initial_list)+1:2*len(initial_list)+1])**2)
        #In the third stage, wavelength, intensity_each, CagliottiUVW, bckg_poly,
        # tube_sample_distances, 2theta_bank_shifts, tube_perpendicular_shifts are refined
        parameters2 = []
        parameters2.extend(result.x[:len(initial_list)+1])
        parameters2.extend(fwhm_par)
        parameters2.extend(result.x[2*len(initial_list)+1:])
        result = minimize(system_min_expression, parameters2, args=(poly_deg,\
            invd2_list[:len(initial_list)], len(tube_sample_distances_mm),\
                 len(height_above_plane_mm), len(ttheta_shifts),\
                    tube_sample_distance_indices[detector_indx2fit],\
                        tube_sample_shift[detector_indx2fit],\
                            tube_height_indices[detector_indx2fit],\
                                tube_height_multiplicator[detector_indx2fit],\
                                    ttheta_shift_indices[detector_indx2fit],\
                                        ch2pos_restricted, self.center[detector_indx2fit],\
                                            limits_restricted,\
                                                self.ref_angles, x_data, y_data, dy_data, False,\
                                                    None, 0.0, height_above_plane_mm),\
                                                        method='Nelder-Mead',\
                                                            options={'maxiter':2000000000})
        print "2/4 has been finished(Cagliotti). Chi^2: ", result.fun
        result = minimize(system_min_expression, result.x, args=(poly_deg,\
            invd2_list[:len(initial_list)], len(tube_sample_distances_mm),\
                 len(height_above_plane_mm), len(ttheta_shifts),\
                    tube_sample_distance_indices[detector_indx2fit],\
                        tube_sample_shift[detector_indx2fit],\
                            tube_height_indices[detector_indx2fit],\
                                tube_height_multiplicator[detector_indx2fit],\
                                    ttheta_shift_indices[detector_indx2fit],\
                                        ch2pos_restricted, self.center[detector_indx2fit],\
                                            limits_restricted,\
                                                self.ref_angles, x_data, y_data, dy_data, False,\
                                                    None, 0.0, height_above_plane_mm),\
                                                        method='Nelder-Mead',\
                                                            options={'maxiter':2000000000})
        print "3/4 has been finished(Cagliotti). Chi^2: ", result.fun
        result = minimize(system_min_expression, result.x, args=(poly_deg,\
            invd2_list[:len(initial_list)], len(tube_sample_distances_mm),\
                 len(height_above_plane_mm), len(ttheta_shifts),\
                    tube_sample_distance_indices[detector_indx2fit],\
                        tube_sample_shift[detector_indx2fit],\
                            tube_height_indices[detector_indx2fit],\
                                tube_height_multiplicator[detector_indx2fit],\
                                    ttheta_shift_indices[detector_indx2fit],\
                                        ch2pos_restricted, self.center[detector_indx2fit],\
                                            limits_restricted,\
                                                self.ref_angles, x_data, y_data, dy_data, False,\
                                                    None, 0.0, height_above_plane_mm),\
                                                        method='Nelder-Mead',\
                                                            options={'maxiter':2000000000})
        print result.message
        print "The chi^2 reached: ", result.fun
        parameters = result.x
        wavelength = parameters[0]
        print "Refined wavelength: ", wavelength
        intensity = np.zeros_like(initial_list)
        print "Intensities: 2*theta, Q, intensity"
        for i, _ in enumerate(initial_list):
            intensity[i] = parameters[1+i]
            print "\t"+str(np.degrees(np.arccos(1.0-0.5*invd2_list[i]*wavelength**2)))+" "+\
                str(np.sqrt(invd2_list[i])*2*pi)+" "+str(intensity[i])
#        fwhm = np.array(result.x[1+len(initial_list):1+2*len(initial_list)])
#        intensity = np.array(result.x[1:1+len(initial_list)])
        fwhm_par = parameters[1+len(initial_list):4+len(initial_list)]
#        fwhm_par = polyn.polyfit(np.tan(\
#            np.arccos(1.0-0.5*invd2_list[:len(initial_list)]*wavelength**2)/2.), fwhm**2, 2,\
#                w=(intensity/fwhm)**2 )
        print "Cagliotti U, V, W: ", fwhm_par[::-1]
        fwhm = polyn.Polynomial(fwhm_par)
#        actual = len(initial_list)+1+len(fwhm)
        actual = len(initial_list)+1+len(fwhm_par)
        poly_raw = np.zeros(poly_deg+1)
        for i in range(poly_deg+1):
            poly_raw[i] = parameters[actual+i]
        #poly = np.poly1d(poly_raw[::-1])
        poly = polyn.Polynomial(poly_raw)
        print "Polynomial background coefficients: ", poly
        actual += poly_deg+1
        for i, _ in enumerate(tube_sample_distances_mm):
            tube_sample_distances_mm[i] = parameters[actual+i]
            print "Tube sample distance no ", i+1, "is ", tube_sample_distances_mm[i]
        actual += len(tube_sample_distances_mm)
        #for i, _ in enumerate(height_above_plane_mm):
        #    height_above_plane_mm[i] = parameters[actual+i]
        #actual += len(height_above_plane_mm)
        for i, _ in enumerate(ttheta_shifts):
            ttheta_shifts[i] = parameters[actual+i]
            print "2Theta shift for bank no ", i+1, "is ", ttheta_shifts[i]
        actual += len(ttheta_shifts)
        difference = np.zeros(len(detector_indx2fit))
        for i, val in enumerate(detector_indx2fit):
            if val:
                difference[i] = parameters[actual]
                actual += 1
                print "2Theta zero for detector no. ", i+1, " is ",\
                    ttheta_shifts[ttheta_shift_indices[i]],\
                        ", perpendicular-tube centre distance is ", difference[i]
        for det in range(self.no_detector_tubes):
            if detector_indx2fit[det]:
                for position, _ in enumerate(self.ref_angles):
                    data = np.loadtxt("nickel_corr_"+str(position+1)+"_"+str(det+1)+".dat",\
                        unpack=True)
                    degrees = calc_degree_from_channel(data[0],\
                        self.ch2pos[det], difference[det],\
                            tube_sample_distances_mm\
                                [tube_sample_distance_indices[det]]+tube_sample_shift[det],\
                                    height_above_plane_mm[tube_height_indices[det]]*\
                                        tube_height_multiplicator[det],\
                                            ttheta_shifts[ttheta_shift_indices[det]]+\
                                                self.ref_angles[position], self.center[det],\
                                                    self.limits[det])
                    model_intensity = calc_model_pattern(wavelength,\
                        invd2_list[:len(initial_list)], intensity, fwhm, poly, degrees)
                    dataset = [degrees] + [data[1][self.limits[det][0]-1:self.limits[det][-1]]] +\
                        [model_intensity] +\
                            [data[2][self.limits[det][0]-1:self.limits[det][-1]]] +\
                               [data[0][self.limits[det][0]-1:self.limits[det][-1]]]
                    np.savetxt("nickel_acorr_"+str(position+1)+"_"+str(det+1)+".dat",\
                        np.column_stack(dataset), header=' angle, Corr, model, dCorr, channel')
        print "Writing calibration parameters into \""+out_calibration+"\" file"
        write_calibration_file(out_calibration, tube_sample_distances_mm, height_above_plane_mm,\
            ttheta_shifts, detector_indx2fit, tube_sample_distance_indices, tube_sample_shift,\
                tube_height_indices, tube_height_multiplicator, ttheta_shift_indices,\
                    self.limits, self.center, self.ch2pos, difference, wavelength, fwhm_par[::-1])
        return

def write_calibration_file(filename, distances, heights, angle_shifts, calibrated,\
    distance_indices, distance_shift, height_index, height_mult, angle_shift_index, limits,\
        centre_chn, ch2pos, tube_shift, wavelength, cagliotti):
    "Writes the calibration parameters into filename"
    fil = open(filename, 'w')
    #wavelenght and Cagliotti U V W
    fil.write("%f %f %f %f \t\t\t\t!wavelength[A], Cagliotti U, V, W\n" % (wavelength,\
        cagliotti[0], cagliotti[1], cagliotti[2]))
    #no. of tubes
    fil.write("%d \t\t\t\t\t\t\t\t!# segments\n" % (len(calibrated)))
    #no. distances, heights, angle shifts
    fil.write("%d %d %d \t\t\t\t\t\t\t\t!# common distances, #common heights," % \
        (len(distances), len(heights), len(angle_shifts)))
    fil.write("#common angular shifts\n")
    #distances
    for _, i in enumerate(distances):
        fil.write("%f " % (i))
    fil.write("\t\t\t\t\t\t\t!common sample-detector distances [mm]\n")
    #heights
    for _, i in enumerate(heights):
        fil.write("%f " % (i))
    fil.write("\t\t\t\t\t\t\t!common height parameters above/below the sample center plane [mm]\n")
    #angle shifts
    for _, i in enumerate(angle_shifts):
        fil.write("%f " % (i))
    fil.write("\t\t\t\t\t\t!common center 2theta angles [degrees]\n")
    #monitor deadtime
    fil.write("??\t\t\t\t\t\t\t\t!deadtime of the monitor counter [us]\n")
    #parameters for each tubes
    for i, _ in enumerate(calibrated):
        fil.write("%d %d %f %d %f %d ??" % (calibrated[i], distance_indices[i]+1,\
            distance_shift[i], height_index[i]+1, height_mult[i], angle_shift_index[i]+1))
        if i == 0:
            fil.write("\t\t\t\t\t!For each segments: (0:omit/1:calibrated) distance_index")
            fil.write(" distance_shift[mm] height_index height_multiplicator angular_shift_index")
            fil.write(" deadtime[us]")
        fil.write("\n")
        #limits
        fil.write("\t%d " % (len(limits[i])-1))
        for _, j in enumerate(limits[i]):
            fil.write("%d " % (j))
        #degree of polynomial
        fil.write("%d " % ((ch2pos[i])[0].degree()))
        #centre channel index
        fil.write("%g " % (centre_chn[i]))
        #polynomials
        for _, j in enumerate(ch2pos[i]):
            for k in range(1, (ch2pos[i])[0].degree()+1):
                fil.write("%g " % ((j.coef)[k]))
        #tube centre shift
        fil.write("%g" % (tube_shift[i]))
        if i == 0:
            fil.write("\t\t\t\t\t!#ranges min_channel {limit_chns} max_chn channel2mm_poly_degree")
            fil.write(" reference_channel ch2mm_poly_coeffs_range1{, then for 2..}")
            fil.write(" perpendicular_shifts_from_reference[mm]")
        fil.write("\n")
    fil.close()
    return

def main(argv):
    "Does the fit"
    #TODO: check the consistency
    if len(argv) != 3:
        print "This program performs the final part of calibration. It is supposed, that channel->"
        print "tube position calibration has been already performed using the masking technique"
        print "\tUsage: "+argv[0]+" control_file.name output_calib.filename"
        print "The control file should contain the following items:"
        print "\t#tubes(incl. non-fitted tubes)"
        print "\tcalibration.nm calib_bckg.nm vanadium.nm vanadium_bckg.nm"
        print "\t#distance_sample-tube #height #detector_banks"
        print "\tcommon_distances[mm]"
        print "\tcommon heights[mm]"
        print "\tperpendicular_angle_shifts_from_encoders_for_each_banks[degrees]"
        print "\tfor each tubes: fit(1)/omit(0) distance_indx[1:] distance_shift height_index[1:]"
        print "\t\theight_multiplicator angle_index #ch->det_ranges"
        print "\t\tlimit_distances(min_channel, separator_ch(if ranges>2) max_channel)"
        print "\t\tch->det_poly_degrees central_channel coeffs_range1 coeffs_range2 ..."
        print "\tWavelength[A] #reflections degree_of_pattern_background"
        print "\tLattice_parameters: a b c [Angstroms] alpha beta gamma [degrees]"
        print "\tthen h k l Miller-indices of each reflections"
        exit()
    with open(argv[1], 'r') as fil:
        print "Reading from the \"", argv[1], "\" control file:"
        # #pos and #tubes
        line = fil.next()
        splt = line.split(' ')
        no_tubes = int(splt[0])
        print "Number of tubes: ", no_tubes
        #filenames
        filenames = []
        line = fil.next()
        splt = line.split(' ')
        print "Filenames are ",
        for j in range(4):
            filenames.append(splt[j])
            print "\""+filenames[j]+"\"",
        print
        # #distances #heights #banks
        line = fil.next()
        splt = line.split(' ')
        no_tube_sample_distances = int(splt[0])
        no_heights = int(splt[1])
        no_angles = int(splt[2])
        print "Number of common sample-tube distances: ", no_tube_sample_distances
        print "Number of common height parameters: ", no_heights
        print "Number of common 2theta shifts of each banks: ", no_angles
        #distances
        distances = np.zeros(no_tube_sample_distances)
        line = fil.next()
        splt = line.split(' ')
        for i in range(no_tube_sample_distances):
            distances[i] = float(splt[i])
        print "Common distances are ", distances
        #heights
        heights = np.zeros(no_heights)
        line = fil.next()
        splt = line.split(' ')
        for i in range(no_heights):
            heights[i] = float(splt[i])
        print "Common heights are ", heights
        #angle shifts
        angle_shifts = np.zeros(no_angles)
        line = fil.next()
        splt = line.split(' ')
        for i in range(no_angles):
            angle_shifts[i] = float(splt[i])
        print "Common angle shifts are ", angle_shifts
        #dataset for each tubes
        fit_indicator = np.zeros(no_tubes, dtype=np.bool_)
        distance_index = np.zeros(no_tubes, dtype=np.int_)
        distance_shift = np.zeros(no_tubes)
        height_index = np.zeros(no_tubes, dtype=np.int_)
        height_multiplicator = np.zeros(no_tubes)
        angle_index = np.zeros(no_tubes, dtype=np.int_)
        center_channel = np.zeros(no_tubes)
        ch2pos_polys = []
        lims = []
        for i in range(no_tubes):
            print "Controls for tube "+str(i+1)+" is about to reading:"
            line = fil.next()
            splt = line.split(' ')
            fit_indicator[i] = int(splt[0])
            if fit_indicator[i]:
                print "\tTakes apart in the calibration"
            else:
                print "\tOmitted from calibration"
            distance_index[i] = int(splt[1])-1
            distance_shift[i] = float(splt[2])
            print "\tDistance index: ", distance_index[i]+1, ", corresponds to ",\
                distances[distance_index[i]], "mm with shift: ", distance_shift[i], "mm = ",\
                    distances[distance_index[i]]+distance_shift[i], "mm"
            height_index[i] = int(splt[3])-1
            height_multiplicator[i] = float(splt[4])
            print "\tHeight index: ", height_index[i]+1, ", corresponds to ",\
                heights[height_index[i]], "mm with multiplicator: ", height_multiplicator[i],\
                    " = ", heights[height_index[i]]*height_multiplicator[i], "mm"
            angle_index[i] = int(splt[5])-1
            print "\tAngle shift index: ", angle_index[i]+1, ", corresponds to ",\
                angle_shifts[angle_index[i]], "degrees"
            no_poly_set = int(splt[6])
            print "\tNumber of ch->tube ranges: "+str(no_poly_set)
            limit_channels = np.zeros(no_poly_set+1, dtype=np.int_)
            for j in range(no_poly_set+1):
                limit_channels[j] = int(splt[7+j])
            print "\tValid range: ", limit_channels[0], "...", limit_channels[-1]
            poly_degree = int(splt[8+no_poly_set])
            print "\tch->tube polynomial degree: ", poly_degree
            center_channel[i] = float(splt[9+no_poly_set])
            print "\tCentral channel: ", center_channel[i]
            poly_temp = []
            for j in range(no_poly_set):
                polys = np.zeros(poly_degree+1)
                for k in range(1, poly_degree+1):
                    polys[k] = float(splt[9+no_poly_set+poly_degree*j+k])
                #poly_temp.append(np.poly1d(polys[::-1]))
                poly_temp.append(polyn.Polynomial(polys))
            ch2pos_polys.append(poly_temp)
            lims.append(limit_channels)
            print "\tChannel->position polynomials: "
            for j in range(len(lims[i])-1):
                if j > 0:
                    print '\t', 4*'-', lims[i][j], 4*'-'
                print ch2pos_polys[i][j]
            print
        line = fil.next()
        splt = line.split(' ')
        wavelength = float(splt[0])
        print "Initial wavelength: ", wavelength, "Angstroms"
        no_reflections = int(splt[1])
        print "Number of reflections: ", no_reflections
        bckg_poly_deg = int(splt[2])
        print "Degree of the background polynomial of the pattern: ", bckg_poly_deg
        lattice_params = np.zeros(6)
        line = fil.next()
        splt = line.split(' ')
        for i in range(6):
            lattice_params[i] = float(splt[i])
        print "Lattice parameters: a=", lattice_params[0], ", b=", lattice_params[1], ", c=",\
            lattice_params[2]
        print "\talpha=", lattice_params[3], "deg, beta=", lattice_params[4], "deg, gamma=",\
            lattice_params[5], "deg"
        reflections_list = []
        for i in range(no_reflections):
            line = fil.next()
            splt = line.split(' ')
            item = np.array([int(splt[0]), int(splt[1]), int(splt[2])])
            reflections_list.append(item)
        print "List of reflections to be taken into account: ", reflections_list
    print "Initializing"
    raw = MTESTCalibrationSystem(filenames, lims, ch2pos_polys, center_channel)
    print "The fitting procedure is about to begin..."
    raw.do_calibration(wavelength, lattice_params, reflections_list, bckg_poly_deg, distances,\
        distance_index, distance_shift, heights, height_index, height_multiplicator, angle_shifts,\
            angle_index, fit_indicator, argv[2])
    return

if __name__ == "__main__":
    main(sys.argv)
