#!/usr/bin/python
"This module describes the averaging and normalization of dataset recorded at MTEST"
import sys
import os.path
import numpy as np
from combine_datasets import MTESTPSDData

class MTESTCalibrationParameters(object):
    "This class provides the calibration parameters of MTEST"
    def __init__(self, filename):
        "Reads in calibration parameters from file"
        fil = open(filename, "r")
        #lambda, Cagliotti UVW
        line = fil.next().lstrip()
        splt = line.split(' ')
        self.wavelength = float(splt[0])    #wavelength in Angstroms
        cagli = np.zeros(3)
        for i in range(3):
            cagli[i] = float(splt[1+i])
        self.cagliotti = np.poly1d(cagli)   #Cagliotti U, V, W parameters: here in polynomial form
        # #segments
        line = fil.next().lstrip()
        splt = line.split(' ')
        no_segments = int(splt[0])
        # #common_distances #common_heights
        line = fil.next().lstrip()
        splt = line.split(' ')
        self.distances = np.zeros(int(splt[0])) #common distances in mm
        self.heights = np.zeros(int(splt[1])) #common heights in mm
        self.ang_shifts = np.zeros(int(splt[2])) #angular shifts in degrees
        #distances
        line = fil.next().lstrip()
        splt = line.split(' ')
        for i, _ in enumerate(self.distances):
            self.distances[i] = float(splt[i])
        # heights
        line = fil.next().lstrip()
        splt = line.split(' ')
        for i, _ in enumerate(self.heights):
            self.heights[i] = float(splt[i])
        # heights
        line = fil.next().lstrip()
        splt = line.split(' ')
        for i, _ in enumerate(self.ang_shifts):
            self.ang_shifts[i] = float(splt[i])
        self.deadtimes = np.zeros(no_segments+1)                #deadtimes of the monitor counters
                                                                #then the detectors
        self.fit_indx = np.zeros(no_segments, dtype=np.bool_)   #0: omit/1: calibrated segment
        self.dist_indx = np.zeros(no_segments, dtype=np.int_)   #distance index of each segment
        self.dist_shft = np.zeros(no_segments)                  #distance shift of each segment
        self.height_indx = np.zeros(no_segments, dtype=np.int_) #height index of each segment
        self.height_mult = np.zeros(no_segments)                #height multiplicator
        self.ang_idx = np.zeros(no_segments, dtype=np.int_)     #angular shift index
        self.limits = []                                        #limit points: min, [barrier], max
        self.reference_channel = np.zeros(no_segments)          #reference channel
        self.ch2mm_poly = []                                    #channel->mm polynomials
        self.perp_shift = np.zeros(no_segments)                 #perpendicular position shift from
                                                                #the reference
        line = fil.next().strip()
        splt = line.split()
        self.deadtimes[0] = float(splt[0])                      #deadtime of the monitor counter
        for i in range(no_segments):
            line = fil.next().lstrip()
            splt = line.split()
            self.fit_indx[i] = int(splt[0])
            self.dist_indx[i] = int(splt[1])-1
            self.dist_shft[i] = float(splt[2])
            self.height_indx[i] = int(splt[3])-1
            self.height_mult[i] = float(splt[4])
            self.ang_idx[i] = int(splt[5])-1
            self.deadtimes[i+1] = float(splt[6])
            line = fil.next().strip()
            splt = line.split()
            no_range = int(splt[0])
            lims = np.zeros(no_range+1)
            for j, _ in enumerate(lims):
                lims[j] = int(splt[1+j])
            self.limits.append(lims)
            actual = 2+no_range
            ch2mm_deg = int(splt[actual])
            self.reference_channel[i] = float(splt[actual+1])
            polys = []
            actual += 2
            for j in range(len(self.limits[-1])-1):
                poly_coeffs = np.zeros(ch2mm_deg+1)
                for k in range(1, ch2mm_deg+1):
                    poly_coeffs[k] = float(splt[actual])
                    actual += 1
                polys.append(np.poly1d(poly_coeffs))
            self.ch2mm_poly.append(polys)
            self.perp_shift[i] = float(splt[actual])
        return

    def get_st_distance(self):
        "Returns with an array containing sample-tube distances"
        dists = np.copy(self.dist_shft)
        for i, _ in enumerate(dists):
            dists[i] += self.distances[self.dist_indx[i]]
        return dists

    def get_vert_pos(self):
        "Returns with an array containing sample-tube distances"
        verticals = np.copy(self.height_mult)
        for i, _ in enumerate(verticals):
            verticals[i] *= self.heights[self.height_indx[i]]
        return verticals

    def get_refang_shft(self):
        "Returns with an array containing shift of reference angles"
        shifts = np.zeros_like(self.height_mult)
        for i, _ in enumerate(shifts):
            shifts[i] = self.ang_shifts[self.ang_idx[i]]
        return shifts

    def get_no_tubes(self):
        "Returns with the number of tubes"
        return len(self.fit_indx)

def main(argv):
    "Test code to create normalized datasets"
    if len(argv) < 5:
        print "This program creates averaged normalized datsets recorded at"+\
            " the MTEST diffractometer"
        print "Usage: "+argv[0]+" outfile calibration.file [-/normalisation"+\
        " _N_ndt_efficiency.dat] {rawfile_names} "
        exit()
    else:
        outfile_name = argv[1]
        print "Name of the output file: \""+outfile_name+"\""
        calibfile_name = argv[2]
        print "Name of the calibration file: \""+calibfile_name+\
            "\" -- reading the calibration parameters.."
        #reading calibration parameters from file
        calib = MTESTCalibrationParameters(calibfile_name)
        print "Calibration parameters have been reading"
        print "Number of datasets: ", str(len(argv)-4)
        normfile = None
        detector_efficiencies = []
        if argv[3] != '-':
            normfile = argv[3]
            print "Each intensities will be normalized by data provided in \""+normfile+\
                "_N_ndt_efficiency.dat\" file"
            i = 1
            while os.path.isfile(normfile+"_"+str(i)+"_ndt_efficiency.dat"):
                normalizer = np.loadtxt(normfile+"_"+str(i)+"_ndt_efficiency.dat", unpack=True)
                detector_efficiencies.append(normalizer[1])
                i += 1
            print("\t"+str(i-1)+" efficiency files have been found.")
        else:
            print "As normalization file is not provided intensities will be kept as original"
            for i in range(calib.get_no_tubes()):
                detector_efficiencies.append(np.ones(1024))
        rawfiles = []
        for i in range(len(argv)-4):
            rawfiles.append(argv[4+i])
            print "\tno. "+str(i+1)+" file is \""+rawfiles[-1]+"\""
        raw = MTESTPSDData(calib.wavelength, '2Theta', rawfiles, calib.limits, calib.ch2mm_poly,\
            calib.get_vert_pos(), calib.get_st_distance(), calib.perp_shift,\
                calib.get_refang_shft(), calib.reference_channel, detector_efficiencies,\
                    calib.deadtimes, 0.02)
        raw.write_summarized2file(outfile_name)
        #raw.write_summarized()
        raw.plot()
    return

if __name__ == "__main__":
    main(sys.argv)
